from functools import reduce
from typing import TypeVar, Callable, List, Set, Generic, Dict, Iterable, Optional, Any
from itertools import islice, chain

T = TypeVar('T')
R = TypeVar('R')
K = TypeVar('K')
U = TypeVar('U')


class Stream(Generic[T]):
    def __init__(self, stream: Iterable[T]):
        self._stream = iter(stream)

    def __iter__(self):
        return self._stream

    @staticmethod
    def of(*args: T) -> 'Stream[T]':
        return Stream(args)

    def map(self, func: Callable[[T], R]) -> 'Stream[R]':
        return Stream(map(func, self._stream))

    def flat_map(self, func: Callable[[T], 'Stream[R]']) -> 'Stream[R]':
        return Stream(chain.from_iterable(map(func, self._stream)))

    def filter(self, func: Callable[[T], bool]) -> 'Stream[T]':
        return Stream(filter(func, self._stream))

    def for_each(self, func: Callable[[T], None]) -> None:
        for i in self._stream:
            func(i)

    def distinct(self):
        return Stream(list(dict.fromkeys(self._stream)))

    def sorted(self, key=None, reverse=False) -> 'Stream[T]':
        return Stream(sorted(self._stream, key=key, reverse=reverse))

    def count(self) -> int:
        return len(list(self._stream))

    def sum(self) -> 'T':
        return sum(self._stream)

    def group_by(self, classifier: Callable[[T], K]) -> Dict[K, List[T]]:
        groups = {}
        for i in self._stream:
            groups.setdefault(classifier(i), []).append(i)
        return groups

    def reduce(self, func: Callable[[T, T], T], initial: T = None) -> 'T':
        if initial:
            return reduce(func, self._stream, initial)
        else:
            return reduce(func, self._stream)

    def limit(self, max_size: int) -> 'Stream[T]':
        return Stream(islice(self._stream, max_size))

    def skip(self, n: int) -> 'Stream[T]':
        return Stream(islice(self._stream, n, None))

    def min(self, key: Callable[[T], Any] = None, default: T = None) -> T:
        """
        :param default: use default value when stream is empty
        :param key: at lease supported __lt__ method
        """
        if key is not None:
            return min(self._stream, key=key, default=default)
        return min(self._stream)

    def max(self, key: Callable[[T], Any] = None, default: T = None) -> T:
        """
        :param default: use default value when stream is empty
        :param key: at lease supported __lt__ method
        """
        if key is not None:
            return max(self._stream, key=key, default=default)
        return max(self._stream)

    def find_first(self) -> Optional[T]:
        try:
            return next(self._stream)
        except StopIteration:
            return None

    def any_match(self, func: Callable[[T], bool]) -> bool:
        """
        this is equivalent to
            for i in self._stream:
                if func(i):
                    return True
            return False
        :param func:
        :return:
        """
        return any(map(func, self._stream))

    def all_match(self, func: Callable[[T], bool]) -> bool:
        return all(map(func, self._stream))

    def none_match(self, func: Callable[[T], bool]) -> bool:
        return not self.any_match(func)

    def to_list(self) -> List[T]:
        return list(self._stream)

    def to_set(self) -> Set[T]:
        return set(self._stream)

    def to_map(self, k: Callable[[T], K], v: Callable[[T], U]) -> Dict[K, U]:
        return {k(i): v(i) for i in self._stream}



import re

import nonebot
from nonebot import Driver, logger
from nonebot import on_command
from nonebot.adapters.onebot.v11 import GroupMessageEvent, Message, Bot, ActionFailed
from nonebot.exception import FinishedException
from nonebot.matcher import Matcher
from nonebot.params import CommandArg, ArgPlainText, ArgStr
from nonebot.permission import SUPERUSER
from nonebot.typing import T_State

from .psql_db import db_context
from .service import game_handler, player_handler, skill_handler, battle_handler, adv_handler
from .service.adv_handler import get_user_status, adv_time_pass, go_outside
from .service.battle_handler import query_battle_log, foo
from .service.forge_handler import get_forge_list, handle_forge, get_forge_num
from .service.game_handler import load_world_data, WorldInfo, Compose
from .service.look_handler import handle_look_all
from .service.mission_handler import get_available_mission, handle_receive_mission, get_submitable_mission, \
    handle_submit_mission, handle_look_mission, handle_del_mission, get_short_submitable_mission
from .service.player_handler import register_new_player, get_player, get_user_status_str, handle_set_stay, \
    handle_show_bag
from .model.player_model import PlayerDB
from .service.shop_handler import get_store_list, buy_item_handle
from .service.skill_handler import get_skill_list, check_skill_exist, get_equip_skill_list, unequip_skill
from .service.use_handler import get_usable_item, use_item_handler
from .utils import get_image, send_group_msg, send_group_msg2, send_group_msg_pic, send_group_msg_pic2, send_img
from .service.help_handler import handle_help, get_msg

register = on_command("加入只因进化录", aliases={"注册只因进化录"}, priority=5, block=True)
set_out = on_command("探索", aliases={"出发"}, priority=5, block=True)
go_home = on_command("返回", aliases={"归来", "回家", "只因返回"}, priority=5, block=True)
force_go_home = on_command("强行返回", aliases={"强制返回"}, priority=5, block=True)
use_item = on_command("只因使用", aliases={"使用", "使用物品"}, priority=5, block=False)
receive_mission = on_command("领取任务", aliases={"接取任务"}, priority=5, block=True)
submit_mission = on_command("提交任务", aliases={"任务完成", "提价任务", "交任务", "任务提交", "完成任务"}, priority=5, block=True)
look_mission = on_command("查看任务", aliases={"我的任务", "查询任务"}, priority=5, block=True)
del_mission = on_command("删除任务", priority=5, block=True)
skill_manager = on_command("技能管理", priority=5, block=False)
equip_skill = on_command("装备技能", aliases={"装备技能", "技能装备", "装备", "只因装备"}, priority=5, block=False)
look_skill = on_command("查看技能", aliases={"查询技能", "我的技能", "技能查看", "查询技能", "只因技能"}, priority=5, block=False)
remove_skill = on_command("移除技能", aliases={"卸载技能", "卸除技能", "删除技能", "移除", "卸载"}, priority=5, block=False)
query = on_command("查询", aliases={"只因查询", "查看", "只因查看"}, priority=5, block=True)
ikun_help = on_command("只因帮助", aliases={"帮助"}, priority=5, block=True)
stay = on_command("驻留", aliases={"滞留", "驻足", "停留", "驻守"}, priority=5, block=True)
reload_config = on_command("重载配置", priority=5, block=False, permission=SUPERUSER)
show_bag = on_command("只因背包", aliases={"查询背包", "背包"}, priority=5, block=True)
rename = on_command("只因改名", priority=5, block=False)
# 模拟战斗 测试用。[[绝对]]不要把这个功能开放給玩家
sim_battle = on_command("模拟战斗", aliases={"战斗模拟"}, priority=5, block=True, permission=SUPERUSER)
# 给所有人发放1个沙漏(已经被签到代替了，这个功能最好不要使用，因为游戏只能玩一个月，发一次少半天)
reward = on_command("发放奖励", aliases={"发放资源"}, priority=5, block=True, permission=SUPERUSER)
# 下面属于历史残留指令 没有用
my_status = on_command("只因状态", priority=5, block=True, permission=SUPERUSER)
game_store = on_command("只因商店", aliases={"只因进化录商店"}, priority=5, block=True, permission=SUPERUSER)
# equip_item = on_command("装备", priority=5, block=True, permission=SUPERUSER)
test = on_command("测试", priority=5, block=True, permission=SUPERUSER)
forge = on_command("制作", priority=5, block=True, permission=SUPERUSER)

driver: Driver = nonebot.get_driver()

driver.on_startup(db_context.init)
driver.on_shutdown(db_context.disconnect)

@driver.on_startup
async def events_read():
    await load_world_data()


# 注册部分
@register.handle()
async def handle_first_register(event: GroupMessageEvent):
    player = await get_player(event)
    if player:
        await register.reject(f"你已经有账号:{player.name} 了哦")
    user_name = (
        event.sender.card if event.sender.card else event.sender.nickname
    )
    await register.finish(await register_new_player(event, user_name, "原版"))


# 出发历险
@set_out.handle()
async def _(state: T_State, event: GroupMessageEvent, arg: Message = CommandArg(), matcher=Matcher()):
    player = await get_player(event)
    if not player:
        await set_out.finish("你还没有账号，请先输入'加入只因进化录'创建账号！")
    await set_out.send(await player_handler.auto_sign(player))

    msg = arg.extract_plain_text().strip()
    item, tmp = await get_usable_item(event)
    state["item"] = item

    player, status, pos_list = await get_user_status(event)
    state["pos_list"] = pos_list
    state["player"] = player
    if status != "休息中":
        await set_out.finish(f"你{status}，不能再出发了")
    if msg != "":
        matcher.set_arg("pos", Message(msg))
        return
    await set_out.send(user_status_to_str(pos_list))


@set_out.got("pos")
async def _(event: GroupMessageEvent, state: T_State, num: str = ArgStr("pos")):
    num = num.strip()
    if not num.isdigit():
        await set_out.finish("输入的格式不对，请输入数字")
    num = int(num)
    if num == 0:
        return
    if num > len(state["pos_list"]) or num < 0:
        await set_out.finish("输入的数字范围不对")
    pos = state["pos_list"][num - 1]
    await go_outside(event, pos)
    await set_out.send(f"{state['player'].name}去{pos}了！")


@my_status.handle()
async def _(event: GroupMessageEvent):
    player = await get_player(event)
    if not player:
        await my_status.finish("你还没有账号，请先输入'加入只因进化录'创建账号！")
    await my_status.finish(await get_user_status_str(event))


def user_status_to_str(pos_list: list[str]):
    tmp = ""
    for i, s in enumerate(pos_list):
        tmp += f"{i + 1}: {s}\n"
    return f"""你要去哪里?(输入数字):
0: 我不去了
{tmp}"""


@go_home.handle()
async def handle_go_home(bot: Bot, event: GroupMessageEvent, state: T_State):
    player = await get_player(event)
    if not player:
        await go_home.finish("你还没有账号，请先输入'加入只因进化录'创建账号！")

    await go_home.send(await player_handler.auto_sign(player))
    log, flg = await adv_time_pass(event)
    state["player_name"] = player.name
    log_str = [("探索简报", s) for s in log.split("🔚")]
    await super_send(bot, go_home, event, log_str, "消息被风控，可等待一段时间后输入【返回】")


@force_go_home.handle()
async def _(bot: Bot, event: GroupMessageEvent):
    log, flg = await adv_time_pass(event, force_go_home=True)
    log_str = [("探索简报", s) for s in log.split("🔚")]
    try:
        await send_group_msg2(bot, event, log_str)
    except Exception as e:
        await force_go_home.send("消息被风控，尝试10s后重发。如果依然失败，可等待一段时间后输入【查询上次】")
        import time
        time.sleep(10)
        await send_group_msg2(bot, event, log_str)
    if not flg:
        await force_go_home.finish()


@game_store.handle()
async def _(event: GroupMessageEvent, state: T_State):
    tmp, tmpstr = await get_store_list(event)
    await game_store.send(tmpstr)
    state["tmp"] = tmp


@game_store.got("choose", "需要买什么呢？(输入数字编号)")
async def _(state: T_State, p: str = ArgStr("choose")):
    p = p.strip()
    if not p.isdigit():
        await game_store.finish("输入的格式不对，请输入数字")
    p = int(p)
    if p == 0:
        await game_store.finish("你离开了商店")
    if p > len(state["tmp"]) or p < 0:
        await game_store.finish("输入的数字范围不对")
    state["item"] = state["tmp"][p - 1]


@game_store.got("choose2", f"要买几个呢？(输入数量)")
async def _(event: GroupMessageEvent, state: T_State, num: str = ArgStr("choose2")):
    num = num.strip()
    if not num.isdigit():
        await game_store.finish("输入的格式不对，请输入数字")
    num = int(num)
    if num < 0:
        await game_store.finish("输入的格式不对，请输入正数")
    if num == 0:
        await game_store.finish("你离开了商店")

    await game_store.finish(await buy_item_handle(event, state["item"].name, state["item"].cost, num))


@use_item.handle()
async def _(event: GroupMessageEvent, state: T_State, arg: Message = CommandArg(), matcher=Matcher()):
    msg = arg.extract_plain_text().strip()
    item, tmp = await get_usable_item(event)
    state["item"] = item

    if msg != "":
        matcher.set_arg("choose", Message(arg))
        return
    if len(item) == 0:
        await use_item.finish("没有可使用的物品")
    await use_item.send(tmp)


@use_item.got("choose")
async def _(bot: Bot, event: GroupMessageEvent, state: T_State, num: str = ArgStr("choose")):
    num = num.strip()
    if not num.isdigit():
        await use_item.finish("输入的格式不对，请输入数字")
    num = int(num)
    if num == 0:
        await use_item.finish()
    if num < 0 or num > len(state["item"]):
        await use_item.finish("输入的数字不在选定范围内")
    msg = await use_item_handler(event, itemname=state["item"][num - 1][0], itemcnt=1)
    await use_item.send(msg)
    if "完成了探索" in msg:
        await handle_go_home(bot, event, state)


@forge.handle()
async def _(event: GroupMessageEvent, state: T_State):
    item, tmp = await get_forge_list(event)
    state["item"] = item
    await forge.send(tmp)


@forge.got("choose", "需要制作什么呢？(输入数字编号)")
async def _(event: GroupMessageEvent, state: T_State, num: str = ArgStr("choose")):
    num = num.strip()
    if not num.isdigit():
        await forge.finish("输入的格式不对，请输入数字")
    num = int(num)
    if num == 0:
        await forge.finish("好的")
    if num < 0 or num > len(state["item"]):
        await forge.finish("输入的数字不在选定范围内")
    can_forge_num = await get_forge_num(event, state["item"][num - 1])
    if can_forge_num <= 0:
        await forge.finish("你的素材不够制作")
    if can_forge_num == 1:
        await forge.finish(await handle_forge(event, state["item"][num - 1], 1))
    await forge.send(f"看起来最多可以做{can_forge_num}个")
    state["tar"] = state["item"][num - 1]
    state["tarnum"] = can_forge_num


@forge.got("choose2", "需要制作几个呢？(输入数字)")
async def _(event: GroupMessageEvent, state: T_State, num: str = ArgStr("choose2")):
    num = num.strip()
    if not num.isdigit():
        await forge.finish("输入的格式不对，请输入数字")
    num = int(num)
    can_forge_num: int = state["tarnum"]
    tar: Compose = state["tar"]
    if num == 0:
        await forge.finish("你不做了")
    if num > can_forge_num:
        await forge.send(f"没法做那么多！帮你做{can_forge_num}个吧！")
        num = can_forge_num
    await forge.finish(await handle_forge(event, tar, num))


#
# @equip_item.handle()
# async def _(event: GroupMessageEvent, state: T_State):
#     tmp, tmpstr = await get_equip_list(event)
#     if len(tmp) == 0:
#         await equip_item.finish('你没有可用的装备')
#     await equip_item.send(tmpstr)
#     state['tmp'] = tmp
#
#
# @equip_item.got("choose", prompt="请输入要装备物品的编号")
# async def _(event: GroupMessageEvent, state: T_State, num: str = ArgPlainText("choose")):
#     if not num.isdigit():
#         await equip_item.finish("输入的格式不对，请输入数字")
#     num = int(num)
#     if num == 0:
#         return
#     if num < 0 or num > len(state["tmp"]):
#         await equip_item.finish("输入的数字不在选定范围内")
#     await equip_item.finish(await handle_equip(event, state["tmp"][num - 1]))


@test.handle()
async def _(event: GroupMessageEvent, bot: Bot, arg: Message = CommandArg()):
    # await foo()
    await test.finish("什么都没执行")
    return


@receive_mission.handle()
async def _(event: GroupMessageEvent, state: T_State, matcher: Matcher):
    ms, ms_str = await get_available_mission(event)
    ms_received = await get_short_submitable_mission(event)
    if len(ms) == 0:
        await receive_mission.finish(ms_received + "没有可领取的任务，可以尝试【查看任务】")
    if len(ms) == 1:
        matcher.set_arg("choose", Message("1"))
    else:
        await receive_mission.send(ms_received + ms_str)
    state["tmp"] = ms


@receive_mission.got("choose")
async def _(bot: Bot, event: GroupMessageEvent, state: T_State, num: str = ArgPlainText("choose")):
    num = num.strip()
    if not num.isdigit():
        await receive_mission.finish("输入的格式不对，请输入数字")
    num = int(num)
    if num == 0:
        return
    if num < 0 or num > len(state["tmp"]):
        await receive_mission.finish("输入的数字不在选定范围内")
    await send_group_msg2(bot, event, await handle_receive_mission(event, state["tmp"][num - 1]))


@submit_mission.handle()
async def _(event: GroupMessageEvent, state: T_State, matcher: Matcher):
    ms, msstr, flg = await get_submitable_mission(event)
    if len(ms) == 0:
        await submit_mission.finish("没有可提交的任务")
    if not flg:
        await submit_mission.finish(msstr)
    if len(ms) == 1:
        matcher.set_arg("choose", Message("1"))
    else:
        await submit_mission.send(msstr)
    state["tmp"] = ms


@submit_mission.got("choose")
async def _(bot: Bot, event: GroupMessageEvent, state: T_State, num: str = ArgPlainText("choose")):
    num = num.strip()
    if not num.isdigit():
        await submit_mission.finish("输入的格式不对，请输入数字")
    num = int(num)
    if num == 0:
        return
    if num < 0 or num > len(state["tmp"]):
        await submit_mission.finish("输入的数字不在选定范围内")
    result = await handle_submit_mission(event, state["tmp"][num - 1])
    if type(result) == str:
        await submit_mission.finish(result)
    await send_group_msg2(bot, event, result)


@look_mission.handle()
async def _(event: GroupMessageEvent, state: T_State, matcher: Matcher):
    ms, msstr, flg = await get_submitable_mission(event)
    if len(ms) == 0:
        await look_mission.finish("没有可查看的任务")
    if len(ms) == 1:
        matcher.set_arg("choose", Message("1"))
    else:
        await look_mission.send(msstr)
    state["tmp"] = ms


@look_mission.got("choose")
async def _(event: GroupMessageEvent, bot: Bot, state: T_State, num: str = ArgPlainText("choose")):
    num = num.strip()
    if not num.isdigit():
        await look_mission.finish("输入的格式不对，请输入数字")
    num = int(num)
    if num == 0:
        return
    if num < 0 or num > len(state["tmp"]):
        await look_mission.finish("输入的数字不在选定范围内")
    await send_group_msg2(bot, event, await handle_look_mission(state["tmp"][num - 1]))


@del_mission.handle()
async def del_mission_handle(event: GroupMessageEvent, arg: Message = CommandArg()):
    msg = arg.extract_plain_text().strip().split()
    if len(msg) < 1:
        await del_mission.finish("参数不完全，请输入删除任务 真实编号 注意，该指令非常危险！")
    id = int(msg[0])
    await del_mission.send(await handle_del_mission(event, id))


@skill_manager.handle()
async def skill_manager_handle(bot: Bot, event: GroupMessageEvent, matcher: Matcher,
                               args: Message = CommandArg()):
    await skill_manager.finish("这个功能已经被关闭了！请使用【查看技能】【装备技能】代替吧")
    msg: list[str] = args.extract_plain_text().strip().split(" ")
    data = dict(enumerate(msg))
    if data.get(0):
        matcher.set_arg("choose", Message(data[0]))
    else:
        await send_group_msg2(bot, event, await get_skill_list(event))


@skill_manager.got("choose")
async def skill_manager_step2(state: T_State, num: str = ArgPlainText("choose")):
    if not num.isdigit():
        await skill_manager.finish("已退出技能管理")
    num = int(num)
    if num == 0:
        await skill_manager.finish()
    if num < 0 or num > 4:
        await skill_manager.finish("输入的数字不在选定范围内")
    state["choose"] = num
    if num == 1:
        await skill_manager.send("请输入你需要装备的技能名")
        return
    if num == 2:
        await skill_manager.send("请输入你需要卸除的技能名")
        return
    await skill_manager.reject("请输入正确选项")


@skill_manager.got("choose2")
async def skill_manager_step3(event: GroupMessageEvent, bot: Bot, state: T_State,
                              choose: str = ArgPlainText("choose2")):
    state["choose2"] = choose
    if state["choose"] == 1:
        result, msg = await check_skill_exist(event, choose)
        if not result:
            await skill_manager.finish(msg)
        state["name"] = msg
        await send_group_msg2(bot, event, await get_equip_skill_list(event, "请选择你要装备的技能位置"))

    if state["choose"] == 2:
        await skill_manager.finish(await unequip_skill(event, choose))


@skill_manager.got("choose3")
async def skill_manager_step3(event: GroupMessageEvent, state: T_State,
                              choose: str = ArgPlainText("choose3")):
    choose = int(choose)
    await skill_manager.finish(await skill_handler.equip_skill(event, state["choose2"], choose - 1))


@ikun_help.handle()
async def show_help_menu(bot: Bot, event: GroupMessageEvent, matcher: Matcher, args: Message = CommandArg()):
    player = await get_player(event)
    if not player:
        await set_out.finish("你还没有账号，请先输入'加入只因进化录'创建账号！")
    tmp: list[str] = args.extract_plain_text().strip().split(" ")
    if len(tmp) == 1 and tmp[0] != "":
        matcher.set_arg("choose", args)
        return
    msg = get_msg()
    # 4. 什么是地图热度？
    await send_group_msg2(bot, event, msg)


@ikun_help.got("choose")
async def show_help_menu_step2(event: GroupMessageEvent, bot: Bot, num: str = ArgPlainText("choose")):
    await handle_help(num, ikun_help, bot, event)


@reload_config.handle()
async def reload():
    await game_handler.load_world_data()
    await reload_config.finish("重载完毕，请注意控制台有无错误")


# todo 无论查询技能 物品 地图信息 怪物 战斗日志 都可以用这个指令
@query.handle()
async def _(matcher: Matcher, bot: Bot, event: GroupMessageEvent, args: Message = CommandArg()):
    plain_text = args.extract_plain_text().strip()
    # 查数字一律认为是在查战斗日志
    if plain_text.isdigit():
        await super_send(bot, query, event, await query_battle_log(plain_text), "查询失败，消息被风控")
        await query.finish()
    if plain_text == '战斗次数':
        await query.finish(f"战斗了{await player_handler.query_times(event)}次", at_sender=True)
    if plain_text:
        matcher.set_arg("name", args)


@query.got("name", prompt="请输入查询名词的部分名称")
async def query_step2(bot: Bot, event: GroupMessageEvent, state: T_State, item_name: str = ArgPlainText("name")):
    msg, flg, item = await handle_look_all(event, item_name)
    if flg:
        await super_send(bot, query, event, msg, "查询失败，消息被风控")
        await query.finish()
    else:
        state["item"] = item
        await query.send(msg)


@query.got("choose")
async def query_step3(bot: Bot, state: T_State, event: GroupMessageEvent, num: str = ArgPlainText("choose")):
    num = num.strip()
    if not num.isdigit():
        await query.finish("输入的格式不对，请输入数字")
    num = int(num)
    if num < 0 or num >= len(state["item"]):  # 因为没有0号选项所以num后面是>= 与其他地方不同
        await query.finish("输入的数字不在选定范围内")
    msg = (await handle_look_all(event, state["item"][num]))[0]
    await super_send(bot, query, event, msg, "查询失败，消息被风控")


@stay.handle()
async def _(event: GroupMessageEvent, args: Message = CommandArg()):
    num = args.extract_plain_text().strip()
    if not num.isdigit():  # 不是数字可能是别人在说胡话（
        return
    await query.finish(await handle_set_stay(event, int(num)))


@show_bag.handle()
async def _(event: GroupMessageEvent):
    await show_bag.finish(await handle_show_bag(event))


@equip_skill.handle()
async def _(event: GroupMessageEvent, arg: Message = CommandArg(), matcher=Matcher()):
    msg = arg.extract_plain_text().strip().split(" ")
    if len(msg) == 1 and msg[0] != '':
        # 有的人喜欢不带空格 那就帮他！（e.g. 装备技能 尖角1  装备技能1尖角）
        rr: list[str] = re.split('(\\d+)', msg[0])
        if len(rr) >= 2 and rr[1].isdigit():
            msg = rr[0:2] if rr[0] != '' else rr[1:3]
        else:
            matcher.set_arg("name", Message(msg[0]))
            return

    if len(msg) == 2:
        if msg[0].isdigit() and not msg[1].isdigit():
            msg[0], msg[1] = msg[1], msg[0]
        await equip_skill.finish(await skill_handler.equip_skill(event, msg[0], int(msg[1]) - 1))


@equip_skill.got("name", "你要装备什么技能？（纯汉字）")
async def equip_step2(bot: Bot, state: T_State, event: GroupMessageEvent, name: str = ArgPlainText("name")):
    result, msg = await skill_handler.check_skill_exist(event, name)
    if not result:
        await equip_skill.finish(msg)
    state["name"] = msg
    await send_group_msg2(bot, event, await skill_handler.get_equip_skill_list(event, f"要把{msg}装备在哪里？"))


@equip_skill.got("pos")
async def equip_step3(state: T_State, event: GroupMessageEvent, pos: str = ArgPlainText("pos")):
    await equip_skill.finish(await skill_handler.equip_skill(event, state["name"], int(pos) - 1) +
                             f"\n提示: 可直接输入 装备技能{state['name']}{pos} ，加快操作速度")


@look_skill.handle()
async def _(bot: Bot, event: GroupMessageEvent, arg: Message = CommandArg()):
    msg = arg.extract_plain_text().strip()
    if msg != "":  # 查询技能和查看技能是2个不同的功能 如果玩家混淆了 可能会在查看技能后面加上技能名词 这时候查看技能就不再处理这件事 交给查询技能
        return
    await send_group_msg2(bot, event, (await get_skill_list(event))[1:])


@remove_skill.handle()
async def _(arg: Message = CommandArg(), matcher=Matcher()):
    msg = arg.extract_plain_text().strip().split(" ")
    if len(msg) >= 1 and msg[0] != '':
        logger.info(str(msg), len(msg))
        matcher.set_arg("name", Message(msg[0]))


@remove_skill.got("name", prompt="你想移除的技能名字(汉字),或技能的位置(数字):")
async def remove_skill2(event: GroupMessageEvent, name: str = ArgPlainText("name")):
    await remove_skill.finish(await skill_handler.unequip_skill(event, name))


@reward.handle()
async def handle_reward():
    await reward.finish(await player_handler.reward_all())


@rename.handle()
async def _(matcher: Matcher, args: Message = CommandArg()):
    plain_text = args.extract_plain_text()  # 首次发送命令时跟随的参数，例：/天气 上海，则args为上海
    if plain_text:
        matcher.set_arg("name", args)  # 如果用户发送了参数则直接赋值


@rename.got("name", prompt="你要叫什么名字呢？")
async def handle_rename(event: GroupMessageEvent, name: str = ArgPlainText("name")):
    await rename.finish(await player_handler.change_name(event, name))


@sim_battle.handle()
async def _(bot: Bot, event: GroupMessageEvent, arg: Message = CommandArg()):
    msg = arg.extract_plain_text().strip().split(" ")
    log = await battle_handler.sim_battle(event, msg[0], msg[1], msg[2])
    await send_group_msg2(bot, event, [("最高权限-模拟战斗", log)])


# 支持风控检查的超级发送函数
async def super_send(bot, bot2, event, msg, failmsg, name='', to_pic=False):
    if to_pic:
        if list == type(msg):
            if type(msg[0]) == str:
                await send_group_msg_pic(bot2, name, msg)
                return
            await send_group_msg_pic2(bot2, msg)
            return
        await send_img(bot2, msg)
        return
    try:
        if list == type(msg):
            if len(msg) == 0:
                await bot2.finish("发送消息为空")
            if type(msg[0]) == str:
                await send_group_msg(bot, event, name, msg)
            else:
                await send_group_msg2(bot, event, msg)
        else:
            await bot2.finish(msg)
    except Exception as e:
        if not isinstance(e, FinishedException):
            logger.warning("检查到风控！")
            await bot2.send(failmsg)
            # await super_send(bot, bot2, event, msg, failmsg, name, True)
# @register.got("yn")
# async def handle_register_yn(matcher: Matcher, yn: str = ArgStr("yn")):
#     if yn != "是":
#         await register.finish("重新注册吧！")
#     await register.send(Message.template("{}{}{}{}{}").format(get_image("head", "阿桃", "gif"),
#                                                               get_image("head", "走地鸡", "gif"),
#                                                               get_image("head", "墨镜鸡", "gif"),
#                                                               get_image("head", "原版", "gif"),
#                                                               "选择你的形象(1/2/3/4)"))

#
# @register.got("choose")
# async def handle_register_sex(event: GroupMessageEvent, state: T_State, choose: str = ArgStr("choose")):
#     if not choose.isdigit():
#         await register.reject("请输入正确数字")
#     choose = int(choose)
#     head_name_list = ["阿桃", "走地鸡", "墨镜鸡", "原版"]
#     if 1 <= choose <= 4:
#         await register.finish(await register_new_player(event, state["name"], head_name_list[choose - 1]))
#     else:
#         await register.reject("请输入正确数字")



from .enums import Trigger, DamageType


# 一坨getter和setter 用来保证不会写错变量名
# 虽然写的很挫但是py没有lombok只能这样了
class MsgPack:
    def __init__(self) -> None:
        self.data = {}  # 数据

    # 触发类型
    def trigger_type(self, trigger_type) -> "MsgPack":
        if self.data.get("trigger_type") is not None:
            raise "警告，请勿循环使用同一个pack"
        self.data["trigger_type"] = trigger_type
        return self

    def check_trigger(self, trigger) -> bool:
        return self.data["trigger_type"] == trigger

    # 攻击
    def atk(self, atk) -> "MsgPack":
        self.data["atk"] = atk
        return self

    def get_atk(self) -> int:
        return self.data["atk"]

    def change_atk(self, apply) -> None:
        self.data["atk"] = apply(self.get_atk())

    # 防御
    def defe(self, defe) -> "MsgPack":
        self.data["def"] = defe
        return self

    def get_def(self) -> int:
        return self.data["def"]

    def change_def(self, apply) -> None:
        self.data["def"] = apply(self.get_def())

    # 速度
    def spd(self, spd) -> "MsgPack":
        self.data["spd"] = spd
        return self

    def get_spd(self) -> int:
        return self.data["spd"]

    def change_spd(self, apply) -> None:
        self.data["spd"] = apply(self.get_spd())

    # 生命
    def hp(self, hp) -> "MsgPack":
        self.data["hp"] = hp
        return self

    def get_max_hp(self) -> int:
        return self.data["hp"]

    def change_max_hp(self, apply) -> None:
        self.data["hp"] = apply(self.get_max_hp())

    # 等级
    def lv(self, lv) -> "MsgPack":
        self.data["lv"] = lv
        return self

    def change_lv(self, apply) -> None:
        self.data["lv"] = apply(self.get_lv())

    def get_lv(self) -> int:
        return self.data["lv"]

    # 生命回复速度
    def life_inc_spd(self, spd) -> "MsgPack":
        self.data["life_inc_spd"] = spd
        return self

    def change_life_inc_spd(self, apply) -> None:
        self.data["life_inc_spd"] = apply(self.get_life_inc_spd())

    def get_life_inc_spd(self):
        return self.data["life_inc_spd"]

    # 暴击率
    def crit(self, crit) -> "MsgPack":
        self.data["crit"] = crit
        return self

    def get_crit(self) -> int:
        return self.data["crit"]

    def change_crit(self, apply) -> None:
        self.data["crit"] = apply(self.get_crit())

    # 暴击伤害
    def csd(self, crit) -> "MsgPack":
        self.data["csd"] = crit
        return self

    def get_csd(self) -> int:
        return self.data["csd"]

    def change_csd(self, apply) -> None:
        self.data["csd"] = apply(self.get_csd())

    # 受到的伤害（来自be_attack）
    def damage(self, num) -> "MsgPack":
        self.data["damage"] = int(num)
        return self

    def get_damage(self) -> int:
        return round(self.data["damage"])

    def change_damage(self, apply) -> None:
        self.data["damage"] = apply(self.get_damage())

    # 我方
    def our(self, our: "Pokemon") -> "MsgPack":
        self.data["our"] = our
        self.data["name"] = our.name if our is not None else "无来源"
        return self

    def get_our(self) -> "Pokemon":
        return self.data["our"]

    def get_name(self) -> str:
        return self.data["name"]

    def is_our_owner(self) -> bool:
        return self.data["our"] == self.data["buff_owner"]

    # 敌方
    def enemy(self, enemy) -> "MsgPack":
        self.data["enemy"] = enemy
        return self

    def get_enemy(self):
        return self.data["enemy"]

    # def is_enemy_owner(self) -> bool:
    #     return self.data["enemy"] == self.data["buff_owner"]

    # buff名字（debug用）
    def buff_name(self, buff_name: str):
        self.data["buff_name"] = buff_name
        return self

    def get_buff_name(self):
        return self.data["buff_name"]

    def check_buff_name(self, buff_name: str):
        return buff_name in self.data["buff_name"]

    # 是否禁用后续事件
    def not_allow(self):
        self.data["allow"] = False

    def get_allow(self) -> bool:
        return self.data.get("allow", True)

    # 连锁结算数据包
    def pack(self, pack) -> "MsgPack":
        self.data["pack"] = pack
        return self

    def get_pack(self) -> "MsgPack":
        return self.data["pack"]

    # buff拥有者(后续可能会取代我方)
    def buff_owner(self, buff_owner):
        self.data["buff_owner"] = buff_owner
        return self

    def get_owner(self) -> "Pokemon":
        return self.data["buff_owner"]

    def check_owner(self, owner):
        return owner.id == self.data["buff_owner"].id

    # 开启穿透伤害（来自attack）
    def perfw(self) -> "MsgPack":
        self.data["perfw"] = True
        return self

    def is_perfw(self):
        return self.data.get("perfw", False)

    # 最终伤害系数（来自attack）
    def percent(self, num) -> "MsgPack":
        self.data["percent"] = num
        return self

    def get_percent(self) -> float:
        return self.data.get("percent", 100)

    # 伤害类型
    def damage_type(self, damage_type):
        self.data["damage_type"] = damage_type
        return self

    def check_damage_type(self, type):
        return self.data["damage_type"] == type

    # 受到伤害时，造成伤害的触发器。如果是攻击触发，容易造成反击。毒触发则不会。
    def damage_taken_trigger(self, tri: Trigger):
        self.data["damage_taken_trigger"] = tri
        return self

    def check_damage_taken_trigger(self, tri: Trigger):
        return self.data["damage_taken_trigger"] == tri

    # 回合开始计数
    def turn_count(self, num:int) -> "MsgPack":
        self.data["turn_count"] = num
        return self

    def get_turn_count(self) -> int:
        return self.data["turn_count"]

    def __str__(self) -> str:
        return f"{self.data}"

    @staticmethod
    def builder() -> "MsgPack":
        return MsgPack()

    @staticmethod
    def get_atk_pack():
        return MsgPack.builder().trigger_type(Trigger.GET_ATK)

    @staticmethod
    def get_def_pack():
        return MsgPack.builder().trigger_type(Trigger.GET_DEF)

    @staticmethod
    def get_lv_pack():
        return MsgPack.builder().trigger_type(Trigger.GET_LV)

    @staticmethod
    def get_max_hp_pack():
        return MsgPack.builder().trigger_type(Trigger.GET_HP)

    @staticmethod
    def get_spd_pack():
        return MsgPack.builder().trigger_type(Trigger.GET_SPD)

    @staticmethod
    def get_life_inc_spd_pack():
        return MsgPack.builder().trigger_type(Trigger.GET_LIFE_INC_SPD)

    @staticmethod
    def get_crit_pack():
        return MsgPack.builder().trigger_type(Trigger.GET_CRIT)

    @staticmethod
    def get_csd_pack():
        return MsgPack.builder().trigger_type(Trigger.GET_CSD)

    @staticmethod
    def active_pack():
        return MsgPack.builder().trigger_type(Trigger.ACTIVE)

    @staticmethod
    def atk_pack():
        return MsgPack.builder().trigger_type(Trigger.ATTACK)

    @staticmethod
    def turn_end_pack():
        return MsgPack.builder().trigger_type(Trigger.TURN_END)

    @staticmethod
    def turn_start_pack(num):
        return MsgPack.builder().trigger_type(Trigger.TURN_START).turn_count(num)


    # @staticmethod
    # def be_atk_pack():
    #     return MsgPack.builder().trigger_type(Trigger.BE_ATTACK)

    @staticmethod
    def damage_pack(our, enemy, num, type: DamageType = DamageType.NORMAL):
        return MsgPack.builder().trigger_type(Trigger.DEAL_DAMAGE).damage_type(type).damage(num).our(our).enemy(enemy)

    @staticmethod
    def taken_damage_pack(our, enemy, num, taken_trigger: Trigger, type: DamageType = DamageType.NORMAL):
        return MsgPack.builder().trigger_type(Trigger.TAKEN_DAMAGE).damage_type(type).damage(num).our(our).enemy(enemy) \
            .damage_taken_trigger(taken_trigger)



import math

from nonebot.adapters.onebot.v11 import GroupMessageEvent

from .game_handler import get_world_data, Skill
from ..model.player_model import PlayerDB
from ..utils import get_uid, fill_list


async def get_skill_list(event: GroupMessageEvent):
    player: PlayerDB = await PlayerDB.get_player_by_uid(get_uid(event))
    skill: dict[str, float] = player.skill
    sp_skill: dict[str, float] = player.sp_skill
    # equip: list[str] = player.equip_skill
    max_equip: int = player.system_lv

    # new_equip_str = [f"{eq[0]} 等级{eq[1]}\n" for eq in new_equip]
    world_data = get_world_data()
    skill_str = ''.join(world_data.get_skill(name).format_1_line_detail(lv) for name, lv in skill.items())
    sp_skill_str = ''.join(world_data.get_skill(name).format_1_line_detail(lv) for name, lv in sp_skill.items())

    return [
        (f"超级吞噬系统", "请选择选项：\n1.装备技能\n2.卸除已有技能\n其他.退出技能管理"),
        (f"{player.name}当前装备的技能", equip_to_str(player)),
        ("最大允许装备技能数", str(max_equip)),
        ("技能池", f"{skill_str}"),
        ("血继限界", f"{sp_skill_str}"),
        ("超级吞噬系统", "系统提示，对于不懂的技能·物品均可使用'查询 xx'询问系统。\n可用指令：【装备技能】【查看技能】【移除技能】")
    ]


async def equip_skill(event: GroupMessageEvent, skill_name: str, pos: int):
    # 先检查玩家是否有这个技能
    player: PlayerDB = await PlayerDB.get_player_by_uid(get_uid(event))
    if not (skill_name := player.query_skill_fuzzy(skill_name)):
        return "你还没有这个技能，无法装备"
    # 获取已经装备的技能，因为等级在skill中记录了，所以equip所以是没等级的list而不是skill一样的dict
    equip: list[str] = player.equip_skill
    # 获取总共允许的技能数
    count = player.system_lv
    if pos > count:
        return f"你现在只有{count}个技能槽哦"
    equip = fill_list(equip, count, None)
    # 然后获得已经装备的技能，根据情况分为以下3种
    # 1. 玩家已经装备了A 又装备A在其他位置 这是【换位】
    if skill_name in equip:
        index = equip.index(skill_name)
        equip[index], equip[pos] = equip[pos], equip[index]
        player.equip_skill = equip
        msg = f"交换了{'空闲' if equip[index] is None else equip[index]}与{skill_name}的位置，已装技能：{equip_to_simple_str(player)}"
    # 2. 玩家没有装备A 现在装备A 但是装备的位置已经有B了 这是【替换】
    elif equip[pos] is not None:
        tmp = equip[pos]
        equip[pos] = skill_name
        player.equip_skill = equip
        msg = f"装备了{skill_name}，替换下去了{tmp}，已装技能：{equip_to_simple_str(player)}"
    # 3. 玩家没有装备A 现在装备A 这是普通装备
    else:
        equip[pos] = skill_name
        player.equip_skill = equip
        msg = f"装备了{skill_name}，已装技能：{equip_to_simple_str(player)}"
    await player.update(equip_skill=equip).apply()
    return msg


async def get_equip_skill_list(event: GroupMessageEvent, msg: str):
    player: PlayerDB = await PlayerDB.get_player_by_uid(get_uid(event))
    skill_str = equip_to_str(player)

    return [("超级吞噬系统", msg), ("超级吞噬系统", skill_str)]


def equip_to_str(player: PlayerDB):
    new_equip = [(eq, int(player.skill.get(eq)) if eq is not None else None) for eq in player.equip_skill]
    tmp = ""
    for i in range(player.system_lv):
        if i >= len(new_equip) or new_equip[i][0] is None:
            tmp += f"{i + 1}: (空闲)\n"
        else:
            tmp += f"{i + 1}: {new_equip[i][0]} lv.{new_equip[i][1]}\n"
    return tmp


async def check_skill_exist(event, skill_name):
    player: PlayerDB = await PlayerDB.get_player_by_uid(get_uid(event))
    if skill_name == '吞天者':
        return False, "血继限界是角色天生具备的，不需要装备！"
    if name := player.query_skill_fuzzy(skill_name):
        return True, name
    return False, f"找不到【{skill_name}】这个技能，无法装备"


async def unequip_skill(event, skill_name) -> str:
    player: PlayerDB = await PlayerDB.get_player_by_uid(get_uid(event))

    if skill_name.isdigit():
        i = int(skill_name) - 1
        if i < 0 or i >= player.system_lv:
            return f"输入的数字应当在0-{player.system_lv}之间，已装技能：{equip_to_simple_str(player)}"
    else:
        if skill_name not in player.equip_skill:
            return "你还没有装备这个技能"
        i = player.equip_skill.index(skill_name)

    player.equip_skill[i] = None
    await player.update(equip_skill=player.equip_skill).apply()
    return f"卸除技能完成，已装技能：{equip_to_simple_str(player)}"


def equip_to_simple_str(player: PlayerDB) -> str:
    new_equip = [(eq, int(player.skill.get(eq)) if eq is not None else None) for eq in player.equip_skill]
    tmp = ""
    for i in range(player.system_lv):
        if i >= len(new_equip) or new_equip[i][0] is None:
            tmp += f"(空闲) "
        else:
            tmp += f"{new_equip[i][0]}{new_equip[i][1]} "
    return tmp


def test() -> str:
    log = "防御成长 当前等级 1\n"
    lv = 1
    skill: Skill = get_world_data().get_skill("防御成长")
    for i in range(90):
        lv = skill.skill_lv_up(lv, 5)
        log += f"打母猴{i}次 升级lv.{lv}\n"
    # lv=1
    # for i in range(10):
    #     lv = skill.skill_lv_up(lv, 30)
    #     log += f"打母猴{i}次 升级lv.{lv}\n"
    return log



from nonebot.adapters.onebot.v11 import GroupMessageEvent

from ..psql_db import db
from .achievement_handler import add_achievement
from ..service.adv_handler import get_player_status
from .game_handler import Mission
from .game_handler import get_world_data
from ..model.mission_model import MissionDB
from .player_handler import get_player
from ..model.player_model import PlayerDB
from ..stream import Stream
from ..utils import get_uid


async def get_available_mission(event) -> [list[Mission], str]:
    player: PlayerDB = await get_player(event)
    ms: dict[str, Mission] = get_world_data().get_mission_list()  # 所有任务列表
    am: list[str] = await MissionDB.get_all_received_mission(player.uid)  # 所有已经完成的且只能完成一次的任务的列表

    def checker(mission: Mission):
        if mission.name in am and mission.type2 == '一次' or mission.type2 == '不可领取':
            return False
        return True

    ms2: list[Mission] = Stream(ms.values()).filter(lambda m: checker(m)).to_list()
    tmp = '请选择要领取的任务:\n0: 我不做了\n'
    for i, j in enumerate(ms2):
        tmp += f'{i + 1}: {j.name}\n'
    return ms2, tmp


async def handle_receive_mission(event: GroupMessageEvent, mission: Mission):
    player: PlayerDB = await get_player(event)
    await MissionDB.receive_mission(player.name, get_uid(event), event.group_id, mission.name, mission.type,
                                    mission.type2)
    return [("超级吞噬系统", f"{player.name}成功领取任务：{mission.name}"),
            ("任务描述", mission.des),
            ("任务目标", mission.tar),
            ("任务提示", mission.hint),
            ("任务奖励", mission.reward_str)]


# 返回值 ms 任务列表 tmp 任务描述字符串 flg 领取了重复任务的故障检测
async def get_submitable_mission(event: GroupMessageEvent):
    ms = await MissionDB.get_all_in_progress_mission(get_uid(event))
    tmp = '请选择任务:\n0: 关闭\n'
    # 检查重复元素，防止有人卡bug多次领取任务
    collect = []
    for i, j in enumerate(ms):
        tmp += f'{i + 1}: {j.name}\n'
        collect.append(j.name)
    collect_set = set(collect)
    flg = True
    if len(collect) != len(collect_set):
        tmp += "警告，检测到重复领取相同任务，这可能是系统bug导致的。为了正常游戏体验，请使用删除任务指令删除多余的任务，" \
               "在此之前你将无法提交任务，下面是每个任务对应的‘真实编号’，请使用'删除任务 真实编号'指令清除多余任务，该操作危险，请正确操作。"
        flg = False
        for i, j in enumerate(ms):
            tmp += f'ID {j.id}: {j.name}\n'
    return ms, tmp, flg


async def get_short_submitable_mission(event: GroupMessageEvent):
    ms = await MissionDB.get_all_in_progress_mission(get_uid(event))
    return f"待完成的任务：{', '.join(m.name for m in ms)}\n"


def check_item(item, player):
    for i in item:
        if i["num"] > player.query_item(i["name"]):
            return False
    return True


# def check_equip(equip, player:PlayerDB):
#     for i in equip:
#         if not player.equip.get(i):
#             return False
#     return True


def check_skill(skill, player: PlayerDB):
    for i in skill:
        if i["num"] > player.query_skill(i["name"]):
            return False
    return True


# todo
def check_equip_skill_num(num, player):
    return len(player.equip_skill) >= int(num)


async def handle_submit_mission(event: GroupMessageEvent, ms: MissionDB):
    mission: Mission = get_world_data().get_mission(ms.name)
    player: PlayerDB = await get_player(event)
    # step1 根据时间更新任务事件的变化
    # 还没做
    # step2 检查任务类型
    if item := mission.check.get("check_item"):  # 具备型
        if check_item(item, player):
            return await handle_complete(event, mission, ms, player)
        else:
            return f"任务要求的物品你还没有完成！"

    if skill := mission.check.get("check_skill"):  # 吞食技能型
        if check_skill(skill, player):
            return await handle_complete(event, mission, ms, player)
        else:
            return f"任务要求的技能你还没有取得！"

    if status := mission.check.get("player_status"):  # 状态型
        if await get_player_status(event) == status:
            return await handle_complete(event, mission, ms, player)
        else:
            return f"你还不处于{status}状态哦\n"

    if arrived := mission.check.get("player_arrived"):  # 经历型
        for ar in arrived:
            if player.arrived.get(ar) is None:
                return f"你还没去过{ar}呢！\n"
        return await handle_complete(event, mission, ms, player)

    if require_item := mission.check.get("require_item"):  # 索取型
        if check_item(require_item, player):
            player.cost_items(require_item)
            return await handle_complete(event, mission, ms, player)
        else:
            return f"任务要求的物品你还没有完成！"

    if num := mission.check.get("check_equip_skill_num"):  # 检查装备了几个技能（这种东西真的能复用吗）
        if check_equip_skill_num(num, player):
            return await handle_complete(event, mission, ms, player)
        else:
            return f"你还没有装备上技能！"

    if monster_list := mission.check.get("monster_killed"):
        if t := check_killed(monster_list, player):
            return f"你还没有击杀{t}呢"
        else:
            return await handle_complete(event, mission, ms, player)
    # if equip := mission.check.get("check_equip"):  # 装备检查型
    #     if check_equip(equip, player):
    #         return await handle_complete(event, mission, ms, player)
    #     else:
    #         return f"任务要求的物品你还没有完成！"

    if mission.check.get("impossible"):
        return "这个任务无法完成"

    return "未知错误..."

    # step3 更新日志 反馈


async def handle_complete(event, mission: Mission, ms, player):
    async with db.transaction():
        ans = []
        if mission.finish_str:
            ans += mission.finish_str + '\n'
        await ms.update(status="已完成").apply()
        ans.append(("超级吞噬系统", f"任务【{mission.name}】完成\n"))
        if len(mission.reward) != 0:
            for i in mission.reward.items():
                if i[0] == '系统等级':
                    player.system_lv = player.system_lv + int(i[1])
                    ans.append(("任务奖励", f"系统等级提升，现在可以装备{player.system_lv}个技能了"))
                elif i[0].startswith('称号'):
                    # 称号系统会记录每个人获得的先后
                    arc = await add_achievement(event, i[0][3:], player.times)
                    ans.append(("解锁成就", f"获得{i[0]}，你是第{arc.order}只解锁该称号的因！"))
                else:
                    ans.append(("任务奖励", f"获得 {i[0]} x{i[1]}"))
                    player.add_item(i[0], int(i[1]))
            await player.update(bag=player.bag, collection=player.collection, system_lv=player.system_lv).apply()
        if next_str := mission.next:
            tmp = await handle_receive_mission(event, get_world_data().get_mission(next_str))
            ans += tmp
        return ans


async def handle_look_mission(ms: MissionDB):
    mission: Mission = get_world_data().get_mission(ms.name)
    return [("任务名称", mission.name), ("任务描述", mission.des), ("任务目标", mission.tar), ("任务提示", mission.hint)]


async def handle_del_mission(event: GroupMessageEvent, id: int):
    mission: MissionDB = await MissionDB.get_mission_by_id(id)
    if mission.uid != get_uid(event):
        return "这个任务不是你的！"
    if mission.status == '已完成':
        return "请勿删除已完成的任务"
    await mission.delete()
    return "删除成功..."


def check_killed(monsterlist: dict[str, int], player: PlayerDB):
    killed = player.monster_killed
    for mon, num in monsterlist.items():
        if killed.get(mon, 0) < num:
            return f"{mon}{num}次"

    return None



from ..psql_db import db

from datetime import datetime
from nonebot import logger



class MissionDB(db.Model):
    __tablename__ = "ikun_mission"

    id: int = db.Column(db.Integer(), primary_key=True)
    # 角色ID
    uid = db.Column(db.String(), nullable=False)
    # 群ID
    group_id = db.Column(db.Integer(), nullable=False)
    # 任务名称
    name = db.Column(db.String(), nullable=False)
    # 什么时候去的
    start_time = db.Column(db.DateTime(), default=datetime.now)
    # 懒更新时间
    update_time = db.Column(db.DateTime(), default=datetime.now)
    # 已经提交的物品
    submit_item = db.Column(db.JSON(), nullable=True, default={})
    # 日志
    log = db.Column(db.String(), nullable=False, default="")
    # 完成状态
    status = db.Column(db.String(), nullable=False, default="")
    # 任务类型(主线 支线等等)
    type = db.Column(db.String(), nullable=False)
    # 任务类型(一次 每日等等)
    type2 = db.Column(db.String(), nullable=False)
    # 缺省字段
    reverse1 = db.Column(db.JSON(), nullable=True, default={})
    reverse2 = db.Column(db.JSON(), nullable=True, default={})
    reverse3 = db.Column(db.JSON(), nullable=True, default={})

    @classmethod
    async def receive_mission(cls, player_name: str, uid: str, group_id: int, name: str, type, type2) -> "MissionDB":
        try:
            async with db.transaction():
                return await cls.create(uid=uid, group_id=group_id, name=name, submit_item={},
                                        log=f"{player_name}开始了这个任务\n", type=type, type2=type2, status="进行中")
        except Exception as e:
            logger.info(f"领取任务出错 {type(e)}: {e}")

    @classmethod
    async def get_mission(cls, uid: str, name: str) -> "MissionDB":
        try:
            async with db.transaction():
                return await cls.query.where((cls.uid == uid) & (cls.name == name)).gino.first()
        except Exception as e:
            logger.info(f"任务获取出错 {type(e)}: {e}")

    @classmethod
    async def get_all_in_progress_mission(cls, uid: str) -> list["MissionDB"]:
        try:
            async with db.transaction():
                return await cls.query.where((cls.uid == uid) & (cls.status == '进行中')).gino.all()
        except Exception as e:
            logger.info(f"任务列表获取出错 {type(e)}: {e}")

    # 获取所有领取过的（包括完成和未完成）的任务的名字，只包括一次性任务，这是用来查重防止反复领取的
    @classmethod
    async def get_all_received_mission(cls, uid: int) -> list[str]:
        try:
            async with db.transaction():
                names = await cls.select('name').where((cls.uid == uid) & (cls.type2 == '一次')).gino.all()
                return map(lambda x: x[0], names)
        except Exception as e:
            logger.info(f"任务列表获取出错 {type(e)}: {e}")

    @classmethod
    async def get_mission_by_id(cls, id: int) -> "MissionDB":
        try:
            async with db.transaction():
                return await cls.query.where(cls.id == id).gino.first()
        except Exception as e:
            logger.info(f"任务列表获取出错 {type(e)}: {e}")



from gino import Gino

# 全局数据库连接对象
from nonebot import logger, get_driver

db = Gino()


async def init():
    # if not bind and (not user and not password and not address and not port and not database):
    #     raise ValueError("数据库配置未填写")
    # i_bind = bind
    global_config = get_driver().config
    i_bind = getattr(global_config, "psql")
    try:
        await db.set_bind(i_bind)
        await db.gino.create_all()
        logger.info(f'Database loaded successfully!')
    except Exception as e:
        raise Exception(f'数据库连接错误.... {type(e)}: {e}')


async def disconnect():
    await db.pop_bind().close()



from .. import send_group_msg2, get_image, send_group_msg
from .game_handler import get_world_data


async def handle_help(num, ikun_help, bot, event):
    if not num.isdigit():
        await ikun_help.finish("输入的格式不对，请输入数字")
    num = int(num)
    help = get_world_data().get_help_answer(num)
    if not help:
        await ikun_help.finish("未找到对应帮助")
    if help.type == "图片型":
        await ikun_help.finish(get_image("notice", help.ans))
        return
    if help.type == "消息型":
        await send_group_msg(bot, event,help.name, help.ans)


def get_msg():
    help = "\n".join(get_world_data().get_help_question())

    return [("超级吞噬系统", "请用数字选择具体帮助："),
            ("超级吞噬系统", help)]




from nonebot import logger
from ..psql_db import db
from datetime import datetime



# 成就表
# 单独放一个表是因为这个表组查询操作（计算序号），不太方便丢在player的json里，干脆抽出来方便扩展
class AchievementDB(db.Model):
    __tablename__ = "ikun_achievement"

    id: int = db.Column(db.Integer(), primary_key=True)
    # 角色ID（同时包含了赛季信息）
    uid = db.Column(db.String(), nullable=False)
    # 创建时间
    start_time = db.Column(db.DateTime(), default=datetime.now)
    # 成就名字
    name: str = db.Column(db.String(), nullable=False)
    # 第几个获得这个成就的序号
    order: int = db.Column(db.Integer(), nullable=False)
    # 赛季
    session: str = db.Column(db.String(), default="S1")
    # 完成时总共探索的次数
    times: int = db.Column(db.Integer(), nullable=False, default="999999")

    @classmethod
    async def record_archive(cls, uid: str, name: str, times: int) -> "AchievementDB":
        # 这里要计算一下是第几个获得的，其实业务逻辑不应该写在这但是我懒了，一个事务完成
        try:
            async with db.transaction():
                # # 查询总共有多少个同名成就
                num = await db.select([db.func.count(AchievementDB.id)]) \
                    .where((cls.session == "S1") & (cls.name == name)).gino.scalar()
                return await cls.create(uid=uid, name=name, order=1 + num, times=times)
        except Exception as e:
            logger.info(f"新建成就出错 {type(e)}: {e}")

    @classmethod
    async def query_all_by_id(cls, qid: int) -> list["AchievementDB"]:
        try:
            async with db.transaction():
                return await cls.query.where(cls.uid == qid).gino.all()
        except Exception as e:
            logger.error(f"查询成就出错 {type(e)}: {e}")

    @classmethod
    async def query_by_id_name(cls, qid, name) -> "AchievementDB":
        try:
            async with db.transaction():
                return await cls.query.where((cls.uid == qid) & (cls.name == name)).gino.first()
        except Exception as e:
            logger.error(f"查询成就出错 {type(e)}: {e}")



from ..model.achievement_model import AchievementDB
from ..utils import get_uid


async def add_achievement(event, name: str, times):
    # 已经拿过的成就不能再拿
    if await AchievementDB.query_by_id_name(get_uid(event), name) is not None:
        return None
    return await AchievementDB.record_archive(get_uid(event), name, times)





import math
import tarfile
import unittest

from nonebot import logger
import json5
import os

from ..stream import Stream


class Map:
    name = '未定义地图'  # 地图名称
    enemy_list = []  # 地图拥有的敌人和概率
    item_list = []  # 地图拥有的道具
    owned_map = []  # 子地图
    require_level = 1  # 地图需要的探索等级
    require_honor = {}  # 地图需要的特别称号
    require_buff = {}  # 地图需要的特别buff
    require_item = {}  # 地图需要的特别物品
    cost = {}  # 地图的消耗
    public = True  # 是否直接可达
    description = "无"

    def get_max_enemy_level(self):
        return max(self.enemy_list, key=lambda l: l["lv"])["lv"]


class Item:
    name = '未定义物品'  # 地图名称
    description = "未填写"
    cost = None  # 如果填写cost 就会被商店出售
    lv = 0
    usable = False
    equip_type = None

    def simple_name(self):
        return f"{self.name}(Lv.{self.lv})"


class Compose:
    name = '未定义合成表'
    lv: 1
    type = '熔炼'
    consume = []
    produce = []


class Mission:
    name = None  # 名称
    type = None  # 类型 主线or支线
    type2: str = "一次"  # 能做几次以及类似特性
    hide_reward: bool = False
    des = '未定义任务描述'
    tar = '未定义任务目标'
    check = {}
    hint = '未定义任务提示'
    reward_str = '未定义任务奖励'
    reward: dict[str, int] = {}
    finish_str = None  # 任务完成后说的话
    next = '下一个任务'


class Monster:
    name: str
    lv: int
    atk: int
    hp: int
    defe: int
    ach: str  # 打败他获得的成就，只能获得一次
    spd: int
    des: str  # 怪物描述
    baselv: int  # 填表时填1级怪 会因为怪物等级不是1级失真 baselv设为5就可以填5级数据了
    skill: list[str]
    sp_skill: list[str]  # 体质 没法被吃掉
    hint: str  # 攻略
    # mod_skill: list[str]  # 补正技能（来自地图 活动等）没法被吃掉


class Skill:
    name: str  # 技能名字
    min: int  # 技能最低等级，获得技能时就在这个等级
    max: int  # 最高等级
    des: str  # 技能描述，玩家用查询就可以看到
    exp: int  # 升满级所需要的exp
    eatable: bool  # 可不可以被主角吃掉，极少数技能无法被吃掉
    emoji: str  # 技能的图标
    sp: bool  # 是否是体质技能

    def lv_to_exp(self, lv: float) -> float:
        total_lv = self.max - self.min
        lv -= self.min
        return lv * lv * self.exp / total_lv / total_lv

    # 输入当前技能等级 与 吃掉怪物的技能等级 输出 吃掉怪物后的技能等级（小数视为升级进度）
    def skill_lv_up(self, lv: float, exp: float) -> float:
        # 技能等级与吞食次数的关系可以粗略参考 技能配置.json5
        total_lv = self.max - self.min
        if total_lv == 0:
            return 1
        if self.exp == 0:
            self.exp = 1

        def exp_to_lv(exp: float) -> float:
            return math.sqrt(exp * total_lv * total_lv / self.exp) + self.min

        exp = (exp - self.min) / total_lv + 1
        exp = 1 if exp <= 1 else exp
        new_lv = exp_to_lv(self.lv_to_exp(lv) + exp)
        new_lv = self.max if new_lv >= self.max else new_lv
        if total_lv <= 5:  # 这种技能升级缓慢 2位小数无法准确标示精确粒度
            return round(new_lv, 3)
        return round(new_lv, 2)

    # 获取到下一级还需要多少经验值
    def get_next_lv_exp(self, lv: float) -> float:
        if lv >= self.max:
            return 0
        next_lv = int(lv) + 1
        return round(self.lv_to_exp(next_lv) - self.lv_to_exp(lv), 2)

    # 获取到满级还需要多少经验值
    def get_max_lv_exp(self, lv: float) -> float:
        if lv >= self.max:
            return 0
        next_lv = self.max
        return round(self.lv_to_exp(next_lv) - self.lv_to_exp(lv), 2)

    def format_1_line_detail(self, lv: float):
        if self.sp:
            return f"{self.emoji}{self.name}\n"
        if self.exp == 0:
            return f"{self.emoji}{self.name} 🈵\n"
        tmp = '▁▂▃▄▅▆▇█🈵'
        diff = self.max - self.min
        if diff <= 0:
            percent = '🈵'
        else:
            percent = tmp[int(self.lv_to_exp(lv) * 8) // self.exp]
        return f"{self.emoji}{self.name} lv{lv}/{self.max}{percent}\n"


class Help:
    type: str  # 种类 目前有图片型 消息型2种
    question: str  # 问题
    ans = ""  # 回答 str or list[str]
    name: str  # 消息名称
    pos: int  # 帮助在栏目中的位置


class WorldInfo:
    mapList: dict[str, Map]
    itemList: dict[str, Item]
    forgeList: list[Compose]
    missionList: dict[str, Mission]
    enemyList: dict[str, Monster]
    skillList: dict[str, Skill]
    helpList: dict[int, Help]

    def get_item(self, name) -> Item:
        return self.itemList.get(name)

    def get_map(self, name) -> Map:
        return self.mapList.get(name)

    def get_shop_item(self):  # 如果填写cost 就会被商店出售
        return Stream(self.itemList.items()).map(lambda i: i[1]).filter(lambda i: i.cost).to_list()

    def get_forge_list(self) -> list[Compose]:
        return self.forgeList

    def get_forge(self, name: str):
        return Stream(self.forgeList).filter(lambda f: f.name == name).find_first()

    def get_mission_list(self) -> dict[str, Mission]:
        return self.missionList

    def get_mission(self, name) -> Mission:
        return self.missionList.get(name)

    def get_enemy(self, name) -> Monster:
        return self.enemyList.get(name)

    def get_skill(self, name) -> Skill:
        return self.skillList.get(name)

    def get_help_question(self) -> list[str]:
        return [f"{q.pos}. {q.question}" for q in self.helpList.values()]

    def get_help_answer(self, pos):
        return self.helpList.get(pos)


world_data: WorldInfo


def get_world_data() -> WorldInfo:
    return world_data


######### Extract all files from src_dir to des_dir
def extract_tar_files(src_dir, des_dir):
    files = os.listdir(src_dir)
    for file in files:
        dir_tmp = os.path.join(src_dir, file)
        if not os.path.isdir(dir_tmp):  ##是文件，非文件夹
            # 解压特定文件
            if dir_tmp.endswith("gamedata"):
                # f = zipfile.ZipFile(dir_tmp, mode="r")
                f = tarfile.open(dir_tmp)
                names = f.getnames()
                for name in names:
                    f.extract(name, path=des_dir)
                return
        else:
            extract_tar_files(dir_tmp, des_dir)
    return 0


async def load_world_data() -> None:
    global world_data
    world_data = WorldInfo()
    logger.info(f'【只因进化录】资源载入中')
    path = os.path.dirname(__file__) + '/../gamedata/json/'
    path2 = os.path.dirname(__file__) + '/../gamedata/'
    files = os.listdir(path)
    extract_tar_files(path2, path)
    for file in files:
        try:
            with open(f'{path}{file}', "r", encoding="utf-8") as f:
                if file == '地图.json5':
                    world_data.mapList = parse_map(json5.load(f))
                    logger.info(f'【只因进化录】地图载入完成，共{len(world_data.mapList)}张地图')
                if file == '物品.json5':
                    world_data.itemList = parse_item(json5.load(f))
                    logger.info(f'【只因进化录】物品载入完成，共{len(world_data.itemList)}个物品')
                if file == '制作表.json5':
                    world_data.forgeList = parse_compose(json5.load(f))
                    logger.info(f'【只因进化录】制作表载入完成，共{len(world_data.forgeList)}个配方')
                if file == '任务.json5':
                    world_data.missionList = parse_mission(json5.load(f))
                    logger.info(f'【只因进化录】任务载入完成，共{len(world_data.missionList)}个任务')
                if file == '敌人.json5':
                    world_data.enemyList = parse_enemy(json5.load(f))
                    logger.info(f'【只因进化录】敌人载入完成，共{len(world_data.enemyList)}个敌人')
                if file == '技能配置.json5':
                    world_data.skillList = parse_skill(json5.load(f))
                    logger.info(f'【只因进化录】技能载入完成，共{len(world_data.skillList)}个技能')
                if file == "帮助.json5":
                    world_data.helpList = parse_help(json5.load(f))
                    logger.info(f'【只因进化录】帮助载入完成，共{len(world_data.helpList)}条帮助')
        except Exception as e:
            logger.error(str(e))
            logger.error("读取失败，请修改后使用重载配置命令重新加载！")

    # 校验制作表内的道具是否在物品栏内 稍微有点废资源 问题不大 制作功能暂时关闭了
    # for j in world_data.forgeList:
    #     for i in j.consume:
    #         if get_world_data().get_item(i["name"]) is None:
    #             logger.warning(f"制作表内出现未知道具:{i['name']}")
    #     for i in j.produce:
    #         if get_world_data().get_item(i["name"]) is None:
    #             logger.warning(f"制作表内出现未知道具:{i['name']}")


def parse_map(map_json):
    map_dict = {}
    for key, value in map_json.items():
        mp = Map()
        mp.name = key
        mp.require_level = value["require_level"]
        mp.owned_map = value.get("owned") or []  # 可为空 下同
        mp.enemy_list = value.get("enemy_list", [])
        mp.public = value["public"]
        mp.require_honor = value.get("require_honor", {})
        mp.require_buff = value.get("require_buff", {})
        mp.require_item = value.get("require_item", {})
        mp.item_list = value.get("item_list", {})
        mp.description = value.get("description", None)
        mp.cost = value.get("cost", {})
        map_dict[key] = mp

        if not isinstance(mp.owned_map, list):
            logger.warning(f"地图类型疑似错误：{mp.owned_map}")
        if not isinstance(mp.enemy_list, list):
            logger.warning(f"地图类型疑似错误：{mp.enemy_list}")
    return map_dict


def parse_item(item_json):
    item_dict = {}
    for key, value in item_json.items():
        item = Item()
        item.name = key
        item.description = value["des"]
        item.lv = value["lv"]
        item.usable = value.get("usable")
        item.cost = value.get("cost")
        item.equip_type = value.get('equip_type')
        item.is_book = value.get('is_book')
        item_dict[key] = item

    return item_dict


def parse_compose(compose_json):
    compose_list = []
    for v in compose_json:
        comp = Compose()
        comp.name = v["name"]
        comp.lv = v["lv"]
        comp.type = v.get("type", "熔炼")
        comp.consume = v["consume"]
        comp.produce = v["produce"]
        compose_list.append(comp)

    return compose_list


def parse_mission(mission_json):
    mission_dict = {}
    for v in mission_json:
        mission = Mission()
        mission.name = v["name"]
        mission.type = v["type"]
        if not mission.type in ["主线", "支线", "世界任务"]:
            logger.warning(f"未知任务类型{mission.type}")
        mission.type2 = v.get("type2", "不可领取")  # 多半是某个任务链的后续任务，会在完成前面的任务时自动领取
        mission.des = v["des"]
        mission.tar = v["tar"]
        mission.hint = v["hint"]
        mission.reward = v.get("reward", {})
        mission.reward_str = v.get("reward_str", "未知")
        mission.next = v.get("next")
        mission.check = v["check"]
        mission.hide_reward = v.get("hide_reward")
        mission.finish_str = v.get("finish_str")
        mission_dict[mission.name] = mission

    return mission_dict


def parse_enemy(enemy_json):
    enemy_dict = {}
    for v in enemy_json:
        monster = Monster()
        monster.name = v["name"]
        monster.atk = int(v["atk"])
        monster.defe = int(v["def"])
        monster.spd = int(v["spd"])
        monster.hp = int(v["hp"])
        monster.skill = v["skill"]
        monster.baselv = v.get("baselv", 1)
        monster.des = v.get("des", "作者很懒，还没有为它添加描述")
        monster.sp_skill = v.get("sp", [])
        monster.ach = v.get("ach", None)
        monster.hint = v.get("hint", None)
        enemy_dict[monster.name] = monster

    return enemy_dict


def parse_skill(skill_json):
    skill_dict = {}
    for v in skill_json:
        skill = Skill()
        skill.name = v["name"]
        skill.min = int(v.get("min", 0))
        skill.max = int(v.get("max", 0))
        skill.exp = int(v.get("exp", 0))
        skill.des = v["des"]
        skill.eatable = v.get("eatable", True)
        skill.emoji = v.get("emoji", "❔")
        skill.sp = v.get("sp", False)
        skill_dict[skill.name] = skill

    return skill_dict


def parse_help(help_json):
    help_dict = {}
    i = 1
    for v in help_json:
        help: Help = Help()
        help.type = v["type"]
        help.question = v["question"]
        help.ans = v["ans"]
        help.name = v.get("name", "系统帮助")
        help.pos = v.get("pos", i)
        help_dict[i] = help
        i += 1
    return help_dict


# 跑不起来的测试，气死了
class TestCases(unittest.TestCase):
    def testA(self):
        print("pass")
        self.assertEqual("A", "A")



import datetime
import random
import re
import time
from datetime import datetime

from dateutil.relativedelta import relativedelta
from nonebot.adapters.onebot.v11 import GroupMessageEvent

from ..psql_db import db

from .achievement_handler import add_achievement
from ..service.battle_handler import battle
from ..model.adv_model import ActionDB
from .game_handler import Map, get_world_data, Monster
from ..model.battle_log_model import BattleLogDB
from ..model.player_model import PlayerDB
from ..stream import Stream
from ..utils import add_item, get_uid


async def get_user_can_go(player: PlayerDB):
    names = []
    for key, game_map in get_world_data().mapList.items():
        if game_map.require_level > player.lv or not game_map.public:
            continue
        names.append(key)
    return names


async def get_user_status(event) -> [PlayerDB, list[str]]:
    player = await PlayerDB.get_player_by_uid(get_uid(event))
    status = await get_player_status(event)
    pos_list = await get_user_can_go(player)
    return player, status, pos_list


async def go_outside(event: GroupMessageEvent, pos: str) -> None:
    player: PlayerDB = await PlayerDB.get_player_by_uid(get_uid(event))
    other = {"atk": player.atk, "def": player.defe, "hp": player.hp, "spd": player.spd, "lv": player.lv,
             "skill": player.equip_skill, "sp": player.sp_skill, "base_lv": player.lv}
    t = datetime.now().strftime("%H:%M")
    action = await ActionDB.get_action_by_uid(get_uid(event))
    async with db.transaction():
        if action:
            await action.delete()
        await ActionDB.go_outside(get_uid(event), pos, other, f"🌄({t}){player.name}出发了！前去{pos}🔚")


async def get_player_status(event: GroupMessageEvent) -> str:
    action = await ActionDB.get_action_by_uid(get_uid(event))
    if action is not None and "go_home" not in action.other:
        return f"{action.action}"
    return "休息中"


async def query_status(event: GroupMessageEvent, pos: str) -> None:
    action = await ActionDB.get_action_by_uid(get_uid(event))


async def finish_action(event: GroupMessageEvent, checker) -> bool:
    action: ActionDB = await ActionDB.get_action_by_uid(get_uid(event))
    if not checker(action):
        return False
    # 把任务改成10天前触发的，这样就可以迅速完成了
    await action.update(start_time=action.start_time + relativedelta(days=-10),
                        update_time=action.update_time + relativedelta(days=-10)).apply()
    return True


async def query_last_log(event: GroupMessageEvent):
    action: ActionDB = await ActionDB.get_action_by_uid(get_uid(event))
    if not action:
        return ["你还在出发中"]
    return action.log.split("🔚")


# 计算从出发到现在已经经过了几步(一步为60min)  返回步数,到下一步所需时间
def cal_time(start_time, end_time=None) -> [int, float]:
    end_time = time.mktime(end_time.timetuple()) if end_time else time.time()

    # t = (end_time - time.mktime(start_time.timetuple())) // 2  # debug模式下30秒就有一个步骤
    t = (end_time - time.mktime(start_time.timetuple())) // 3600  # 计算经过了几个时间节点
    t2 = (end_time - time.mktime(start_time.timetuple())) % 3600 / 60  # 离下一个时间节点的时间

    return int(t), round(60 - t2, 1)


def check_access(i, player):
    tmp_map: Map = get_world_data().get_map(i["name"])
    # bag = json.load(player.bag)
    if tmp_map.require_level > player.lv:
        return False
    if tmp_map.require_item:
        for i in tmp_map.require_item.items():
            if player.query_item(i[0]) < i[1]:
                return False
    return True


# 根据角色当前的位置，结合其他信息，给出角色接下来要前去的位置。如果返回与当前位置相同的结果，视为停在原地没有移动
def get_random_pos(gmap: Map, player: PlayerDB):
    # 主逻辑
    tmp = Stream(gmap.owned_map).filter(lambda i: check_access(i, player)).to_list()
    if len(tmp) == 0:
        return gmap
    total = Stream(tmp).map(lambda t: t["weight"]).sum()
    total = total if total > 1 else 1  # 如果total大于1 一定会前往下一个区域 否则可能在原地徘徊
    ran = random.uniform(0, total)
    t: float = 0
    for m in tmp:
        t += m["weight"]
        if t >= ran:
            return get_world_data().get_map(m["name"])
    return gmap


def get_random_enemy(gmap: Map):
    if len(gmap.enemy_list) == 0:
        return None
    tmp = gmap.enemy_list
    total = Stream(tmp).map(lambda t: t["weight"]).sum()
    ran = random.uniform(0, total)
    t: float = 0
    for i in gmap.enemy_list:
        t += i["weight"]
        if t >= ran:
            enemy = get_world_data().get_enemy(i["name"])
            enemy.lv = i["lv"]
            return enemy
    return None


def get_item(gmap: Map):
    if len(gmap.item_list) == 0:
        return None, ""
    else:
        log = ""
        item_list = {}
        for i in gmap.item_list:
            log += f"找到了{i['name']} * {i['count']}\n"
            item_list[i['name']] = i['count']
        return item_list, log


def check_bag(cost, bag):
    for k, v in cost.items():
        if bag.get(k) is None:
            return False
        if bag.get(k) <= v:
            return False
    return True


def cost_bag(cost, bag):
    for k, v in cost.items():
        bag[k] -= v


async def cal_go_home(player: PlayerDB, adv: ActionDB):
    # 结算回家
    player.add_items(adv.item_get)
    await player.update(bag=player.bag, collection=player.collection).apply()
    adv.other["go_home"] = True
    await adv.update(other=adv.other).apply()


def get_skill_exp(our_skill: dict[str, float], enemy_skill: list[str]):
    log = ''
    total_compensate = 0.0
    for skill in enemy_skill:
        skill = re.findall(r'\D+', skill)[0]
        number = re.findall(r'\d+', skill)
        number = 1 if len(number) == 0 else int(number[0])
        tmp = our_skill.get(skill)
        tmp2 = get_world_data().get_skill(skill)

        if tmp is None:
            log += f'🆕习得了新技能{skill},初始等级{tmp2.min}\n'
            our_skill[skill] = tmp2.min
        elif tmp == tmp2.max:
            compensate = 0.08 / len(enemy_skill)
            total_compensate += compensate
            # log += f'🈵{skill}技能已经满级了，本该补偿{round(compensate, 3)}个小沙漏，但是为了控制进度不给了！\n'
            log += f'🈵{skill}技能已经满级了\n'
        else:
            lv = tmp2.skill_lv_up(tmp, number)
            log += f'⏫{skill}技能提升({tmp} → {lv})\n'
            our_skill[skill] = lv
    return our_skill, log, round(total_compensate, 2)

    pass


# 游戏采用懒更新 虽然玩家出发之后感觉每时每刻都在探索，但是实际上所有数据会在【玩家查询】的时刻再计算。
# 所以这个函数叫time_pass 他要计算出这段时间内发生了什么，代码非常非常的复杂，会修改多个数据库，他们被放在一个事务里，以确保一致性
async def adv_time_pass(event: GroupMessageEvent, force_go_home: bool = False) -> [str, bool]:
    # 获取所有需要的数据库（act,player）
    uid = get_uid(event)
    act: ActionDB = await ActionDB.get_action_by_uid(uid)
    if not act:
        return "你还没有出发呢！输入'出发'来出发吧", True
    player: PlayerDB = await PlayerDB.get_player_by_uid(uid)
    pos = get_world_data().get_map(act.position)

    if act.other.get("go_home", False):
        return act.log, True
    # 计算应该继续进行几个时间节点？
    tmp, _ = cal_time(act.start_time, act.update_time)
    already_move = min(10, tmp)  # 已经走的步数
    tmp, next_move_dif = cal_time(act.start_time)
    total_move = min(10, tmp)  # 从开始到现在的总步数
    times = total_move - already_move  # 相减即可获得这次需要更新几步

    log = act.log or ""
    defeated = None
    for i in range(20):  # 视为无限循环
        if i == 0 and already_move == 0:  # 刚出门的地方应该记下来
            add_item(player.arrived, act.position)

        if i >= times:
            break

        # 下一步去哪
        log, pos = find_next_step_pos(act, log, player, pos, not (already_move == 0 and i == 0))

        # 检查消耗
        if not check_bag(pos.cost, player.bag):
            log += f"😫没有以下物品：{pos.cost}，没法坚持下去了\n"
            break
        cost_bag(pos.cost, player.bag)  # 消耗的物品在探险时就已经消耗

        # 检查收获
        monster: Monster = get_random_enemy(pos)
        if monster is not None:
            log += f"⚠野生的{monster.name}(lv.{monster.lv})出现了！\n"
            result, battle_log = battle(act.other, player, monster)
            player.times = player.times + 1
            player.record_monster_meet(monster.name)
            # 记录战斗日志
            logdb = await BattleLogDB.record_battle(uid=get_uid(event),
                                                    our_name=player.name,
                                                    enemy_name=monster.name,
                                                    our_lv=act.other["lv"],
                                                    enemy_lv=monster.lv,
                                                    log={"log": battle_log},
                                                    battle_times=player.times)
            if result == '我方胜利':
                log += f"✌战斗胜利，战斗日志ID:{logdb.id}\n"
                # 战胜了就要拿东西
                if monster.lv >= act.other["lv"]:
                    log += f"🚩战胜强者，等级{act.other['lv']} → {act.other['lv'] + 1}\n"
                    act.other["lv"] = act.other["lv"] + 1
                if monster.ach is not None:
                    if result := await add_achievement(event, monster.ach, player.times):
                        log += f"🔚🏆获得成就【{result.name}】,你是第{result.order}只获得的因！\n"
                player.skill, tmp_log, compensate = get_skill_exp(player.skill, monster.skill)
                log += tmp_log
                player.record_monster_killed(monster.name)
                # player.add_item("小沙漏", compensate)
            else:
                log += f"💀战斗失败💀战斗日志ID:{logdb.id}\n输入'查询 id'可以查看战斗日志\n"
                if monster.hint:
                    meet = player.monster_meet.get(monster.name, 0)
                    killed = player.monster_killed.get(monster.name, 0)
                    d = meet - killed
                    if d < 3:
                        log += f"你已经被{monster.name}打死{d}次了，如果超过3次，就可以通过查询指令查询击败它的攻略\n"
                    if d >= 3:
                        log += f"分析成功，输入”查询{monster.name}可获取击杀攻略“\n"
                else:
                    log += f"这只怪物没有攻略，请自行挑战吧\n"
                defeated = i
                break

        item, _log = get_item(pos)
        log += _log
        if item is not None:
            player.add_items(item)
        log += "🔚"  # 特殊标志用来标记这条消息发送完毕，该换下一条消息了

    async with db.transaction():
        # 计算完毕 增加获得的对应物品 记一下日志
        await act.update(item_get=act.item_get, log=log, position=pos.name,
                         update_time=datetime.now(), other=act.other).apply()
        # 再去掉背包里的相关物品 更新exp和图鉴
        await player.update(bag=player.bag, arrived=player.arrived, skill=player.skill,
                            monster_killed=player.monster_killed, collection=player.collection,
                            times=player.times, monster_meet=player.monster_meet).apply()

        if defeated is not None:
            tmp = times - defeated - 1
            log += f"在你死亡后经过了{tmp}个时间点，作为补偿你获得了{tmp / 10}个沙漏🔚"
            player.add_item("小沙漏", tmp / 10)
            await cal_go_home(player, act)
            return log, True

        if total_move >= 10:  # 时间到了，该返回了
            log += f"☯诅咒的力量发作了。\n你回到了1级（无论如何，探索结束都会回到1级）🔚"
            await cal_go_home(player, act)
            return log, True

        if force_go_home:
            # 根据时间补沙漏
            perk = round((60 - next_move_dif) // 10 * 0.01, 2)
            log += f"强行返回，已经挂机的{round(60 - next_move_dif)}分钟转换为{perk}个小沙漏🔚"
            player.add_item("小沙漏", perk)
            await cal_go_home(player, act)
            return log, True

    return log + f"下一个步骤还要等{next_move_dif}分钟，至多还可挂机{10 - total_move}个步骤，要强行返回的话请输入【强行返回】，要加速完成可使用沙漏", False


def find_next_step_pos(act, log, player, pos, can_stay):
    new_pos = pos
    # 检查去了哪里
    if len(pos.owned_map) == 0 and not act.other.get("is_end", False):
        log += f"🏁已经来到了区域的终点\n"
        act.other["is_end"] = True
    if len(pos.owned_map) >= 1:
        new_pos: Map = get_random_pos(pos, player)
        if bool(new_pos.enemy_list) and new_pos.get_max_enemy_level() >= player.buff.get("stay", 999) and can_stay:
            new_pos = pos  # 下一个地区怪物等级太高，启用驻留
            log += f"⚓驻留中...\n"
            act.other["stay"] = True
        elif new_pos != pos:
            log += f"🐥来到了{new_pos.name}...\n"
        elif act.other.get("stay", False):
            log += f"⚓驻留中...\n"
        else:
            log += f"😵没有找到去下一层的路\n"
    add_item(player.arrived, new_pos.name)
    return log, new_pos


async def should_speed_up_cost(event):
    act = await ActionDB.get_action_by_uid(get_uid(event))
    a, next_move_dif = cal_time(act.start_time)
    total_move = min(10, a)
    if total_move < 10:
        cost = 1 - total_move / 10
        return round(cost - 0.1 + next_move_dif / 600, 2)
    return 0



import math
import random

from nonebot.adapters.onebot.v11 import GroupMessageEvent

from .game_handler import Item
from .game_handler import get_world_data, Compose
from .player_handler import get_player
from ..model.player_model import PlayerDB
from ..stream import Stream


def check(player: PlayerDB, t: Compose) -> bool:
    if player.lv < t.lv:
        return False
    for consume in t.consume:
        if not player.collection.get(consume["name"]):
            return False
    return True


async def get_forge_list(event: GroupMessageEvent) -> [dict[str, Item], str]:
    tmp = get_world_data().get_forge_list()
    player = await get_player(event)
    tmp = Stream(tmp).filter(lambda t: check(player, t)).to_list()
    return tmp, convert_shop_item_to_str(tmp)


def convert_shop_item_to_str(tmp: list[Compose]) -> str:
    ans = "制作列表：\n0: 我不做了 \n"
    i = 1
    for item in tmp:
        ans += f"{i}: {item.name}\n"
        i += 1
    return ans


async def get_forge_num(event, forge: Compose) -> int:
    player: PlayerDB = await get_player(event)
    mi = 999999
    for i in forge.consume:
        mi = min(player.query_item(i["name"]) // i["num"], mi)
    return mi


async def handle_forge(event, forge: Compose, times: int) -> str:
    player: PlayerDB = await get_player(event)
    exp = 0
    for i in forge.consume:
        player.cost_item(i["name"], i["num"] * times)
        if i["name"] == '金币':
            exp += i["num"] *times
    tmpstr = '制作中...\n'
    for i in forge.produce:
        num = i["num"]
        tmp = math.modf(num)
        total = 0
        for j in range(0, times):
            # 如果制作一次得到1.4个 那么40%得到2个 60%得到1个
            num = int(tmp[1])
            if tmp[0] > random.random():
                num += 1
            total += num
            tmpstr += f'{forge.type}了{num}个{i["name"]}\n'
        player.add_item(i["name"], total)
    await player.update(bag=player.bag, collection=player.collection,
                        forge_exp=player.forge_exp + exp).apply()
    return tmpstr



from ..model.player_model import PlayerDB
from .game_handler import get_world_data
from .player_handler import get_player
from ..stream import Stream


async def get_equip_list(event) -> tuple[list[str], str]:
    player = await get_player(event)
    tmp = Stream(player.get_bag().keys()).filter(lambda i: get_world_data().get_item(i).equip_type).to_list()
    tmp_str = '可装备物品列表:\n0: 我不装备了\n'
    i = 1
    for s in tmp:
        tmp_str += f"{i}: {s}\n"
        i = i + 1
    return tmp, tmp_str


async def handle_equip(event, name) -> str:
    player: PlayerDB = await get_player(event)
    item = get_world_data().get_item(name)
    old_item = player.get_equip(item.equip_type)
    if old_item is None:
        player.wear(item.equip_type, item.name)
        await player.update(bag=player.bag, equip=player.equip).apply()
        return f"装备上了{item.simple_name()}"
    else:
        player.unwear(item.equip_type, old_item)
        player.wear(item.equip_type, item.name)
        await player.update(bag=player.bag, equip=player.equip).apply()
        return f"脱下了{old_item},装备上了{item.simple_name()}"



from nonebot.adapters.onebot.v11 import Message

from .player_handler import get_player
from .game_handler import get_world_data
from ..model.mission_model import MissionDB
from ..stream import Stream
from ..utils import get_image


async def handle_look_all(event, name) -> [str, bool, list[str]]:
    player = await get_player(event)
    item_list = Stream(get_world_data().itemList.keys()).filter(lambda n: name in n).to_list()
    skill_list = Stream(get_world_data().skillList.keys()).filter(lambda n: name in n).to_list()
    monster_list = Stream(get_world_data().enemyList.keys()).filter(lambda n: name in n).to_list()
    mission_list = Stream(get_world_data().missionList.keys()).filter(lambda n: name in n).to_list()
    map_list = Stream(get_world_data().mapList.keys()).filter(lambda n: name in n).to_list()

    sum = len(item_list) + len(skill_list) + len(monster_list) + len(mission_list) + len(map_list)
    if sum == 0:
        return "未查到相关信息", True, None

    if (sum == 1 and len(item_list) == 1) or (len(item_list) > 0 and item_list[0] == name):
        return show_item_info(player, item_list[0]), True, None

    if (sum == 1 and len(skill_list) == 1) or (len(skill_list) > 0 and skill_list[0] == name):
        return show_skill_info(player, skill_list[0]), True, None

    if (sum == 1 and len(monster_list) == 1) or (len(monster_list) > 0 and monster_list[0] == name):
        return show_monster_info(player, monster_list[0]), True, None

    if (sum == 1 and len(mission_list) == 1) or (len(mission_list) > 0 and mission_list[0] == name):
        return await show_mission_info(player, mission_list[0]), True, None

    if (sum == 1 and len(map_list) == 1) or (len(map_list) > 0 and map_list[0] == name):
        return show_map_info(player, map_list[0]), True, None

    item_list = item_list + skill_list + monster_list + mission_list + map_list
    return convert_items_to_str_list(item_list), False, item_list


def convert_items_to_str_list(tmp: list[str]) -> str:
    ans = "想查哪个？\n"
    i = 0

    for item in tmp:
        ans += f"{i}: {item}\n"
        i += 1
    return ans


def show_item_info(player, name):
    item = get_world_data().get_item(name)
    return Message.template("{}\n{}\n{}\n获得数量 {}").format(name, item.description,
                                                          get_image("item", f"{name}.png"),
                                                          str(player.collection.get(name, "未获得")))


def show_skill_info(player, name):
    lv = player.skill.get(name)
    skill = get_world_data().get_skill(name)
    if lv is None:
        sp_skill = player.sp_skill.get(name)
        own = "未拥有" if sp_skill is None else f"已拥有"
    else:
        own = f"等级:{lv}"
    msg = f"""{name}  技能等级：[{skill.min}-{skill.max}]
描述：{skill.des}
当前{own}
升级总需经验：{skill.exp}"""
    # 性能开销略大的溯源搜索
    result = []
    for m in get_world_data().mapList.values():
        if player.query_arrived(m.name) == 0:
            continue
        total = sum(enemy["weight"] for enemy in m.enemy_list)
        for enemy in m.enemy_list:
            if enemy["weight"] * 2 < total:
                continue
            for sk in get_world_data().get_enemy(enemy["name"]).skill:
                if name in sk:
                    result.append(f"{enemy['name']}({m.name})")
    if bool(result):
        pos = ", ".join(result)
    else:
        pos = "去过一次才知道"

    if lv is not None:
        next_exp = skill.get_next_lv_exp(lv)
        max_exp = skill.get_max_lv_exp(lv)
        return msg + f"""
升级尚需经验：{next_exp}
满级尚需经验：{max_exp}
已知大概率掉落地点：{pos}"""
    return msg


def show_monster_info(player, name):
    monster = get_world_data().enemyList.get(name)
    meet = player.monster_meet.get(name, 0)
    killed = player.monster_killed.get(name, 0)
    skill = ",".join(monster.skill) if len(monster.skill) != 0 else "无"
    sp_skill = ",".join(monster.sp_skill) if len(monster.sp_skill) != 0 else "无"
    shuxing = f"{monster.baselv}级数值：攻{monster.atk} 防{monster.defe} 命{monster.hp} 速{monster.spd}"
    des = monster.des
    if meet == 0:
        skill = "未知"
        sp_skill = "未知"
        shuxing = "数值：未知"
        des = ""
    if killed == 0 and "反鉴定" in monster.sp_skill:
        skill = "鉴定被干扰！需要击杀怪物"
        sp_skill = "反鉴定，其他未知"
        shuxing = f"数值：未知"
        des = ""
    hint = ""
    if meet < killed:
        hint = "(老怪物遭遇次数在更新前没有统计)"
    if sp_skill == "无" or sp_skill == "未知":
        sp_txt = ""
    else:
        sp_txt = f"血继限界：{sp_skill}"
    d = meet - killed
    if monster.hint:
        if d >= 3:
            hint = "【系统提示】\n"+monster.hint
        else:
            hint = "【系统提示】\n如果你被其击败了3次，就可以看到攻略"
    return f"""{name}
技能：{skill}
{sp_txt}
{shuxing}
遭遇次数：{meet}
击杀次数：{killed}
{des}
{hint}"""


async def show_mission_info(player, name):
    mission = get_world_data().get_mission(name)
    db: MissionDB = await MissionDB.get_mission(player.uid, name)
    if db is None:
        return f"""任务名称：{name}
你还尚未接取过这个任务。"""
    return [("任务名称", name), ("任务描述", mission.des), ("任务目标", mission.tar), ("任务提示", mission.hint),
            ("任务奖励", mission.reward_str), ("完成状态", db.status)]


def show_map_info(player, name):
    m = get_world_data().get_map(name)
    arrived = player.query_arrived(name)
    if arrived == 0:
        return f"地图名称：{name}\n至少去过一次后，展示更多信息"
    sub_map = ", ".join(t["name"] for t in m.owned_map) if len(m.owned_map) != 0 else "无"
    creature = ", ".join(t["name"] for t in m.enemy_list) if len(m.enemy_list) != 0 else "无"
    item = ", ".join(t["name"] for t in m.item_list) if len(m.item_list) != 0 else "无"
    other_info = "" if m.public else f'包含生物：{creature}\n掉落物品：{item}'
    return f"""地图名称：{name}
起点：{"是" if m.public else "否"}
{m.description if m.description else ""}
停留次数：{arrived}
通向哪里：{sub_map}
{other_info}"""



from .logger import Logger
from .tools.stream import Stream
from .buff import Buff, new_buff
from .msgPack import MsgPack
from .enums import Trigger, BuffPriority


class MsgManager:

    def __init__(self, logger: Logger):
        self.logger = logger
        self.buffs: dict[Trigger, list["Buff"]] = {}
        self.force_end = False  # 强制结束游戏
        # self.hooks = []  # 扭曲类效果存放

        self.add_death_watcher()

    def register(self, buff: Buff):
        # self.logger.log(f"新增了buff:{buff}")
        if buff.trigger not in self.buffs:
            self.buffs[buff.trigger] = []
        self.buffs[buff.trigger].append(buff)
        self.buffs[buff.trigger] = sorted(self.buffs[buff.trigger], key=lambda b: b.get_priority())

    def send_msg(self, pack: MsgPack):

        def handle(buff, p):
            if self.force_end:
                return
            if pack.check_trigger(Trigger.ACTIVE):
                if not pack.get_allow():
                    return  # 一个事件一旦被禁止，便无法再被禁止。如果攻击换成了毒，就无法再被换成其他的
            # 打断机制：一个已经附加的buff可能会因为其他buff无法执行（例如毒液攻击会取代普通攻击，这使得普通攻击无法执行）
            # 也就是简化的扭曲机制，把普攻扭曲成了毒液攻击，但是没法做到非常细节的扭曲。
            # 目前只有攻击会被打断，如果每个buff都需要判断是否打断，游戏速度会变慢很多
            if pack.check_trigger(Trigger.ACTIVE) or (not pack.check_trigger(Trigger.ATTACK) and
                                                      not pack.check_trigger(Trigger.DEAL_DAMAGE) and
                                                      not pack.check_trigger(Trigger.TURN_START)):  # 目前只有这3种事件需要active
                DEBUG = None
                if DEBUG:
                    if pack.check_trigger(DEBUG):
                        self.logger.debug_log(f"before{p.get_max_hp()}")
                        print(trigger)

                buff.handle(p)
                if DEBUG:
                    if pack.check_trigger(DEBUG):
                        self.logger.debug_log(f"after{p.get_max_hp()}")

            else:
                _pack: MsgPack = MsgPack.active_pack().pack(p)
                self.send_msg(_pack)
                if _pack.get_allow():
                    buff.handle(p)
                # else:
                #     self.logger.log(f"{buff}被阻止了")

        # if pack.check_trigger(Trigger.TAKEN_DAMAGE):
        #     self.logger.log(f"{pack.get_our()}atk")

        trigger = pack.data["trigger_type"]
        if trigger not in self.buffs:
            return
        for buff in self.buffs[trigger]:
            # self.logger.log(f"{buff} {pack.data['trigger_type']}")
            if not pack.check_trigger(buff.trigger):
                continue
            if not buff.check(pack):
                continue
            handle(buff, pack)

    # 回合结束时需要删除一些过期的buff
    def turn_end(self):
        for k, buff_list in self.buffs.items():
            tmp = []
            for buff in buff_list:
                if buff.get_time() == 9999:
                    tmp.append(buff)
                    continue
                buff.time_pass()
                if buff.get_time() > 0:
                    tmp.append(buff)
                elif buff.get_del_msg():
                    self.logger.log(buff.get_del_msg())
            self.buffs[k] = tmp

    def clean(self):
        self.buffs = {}
        self.add_death_watcher()
        self.hooks = []

    def get_buff_stream(self) -> Stream["Buff"]:
        return Stream(self.buffs).flat_map(lambda k: Stream(self.buffs[k]))

    def add_death_watcher(self):
        # 添加死亡监控
        # 在每次受伤后都进行死亡检测
        def check_death(pack: MsgPack):
            if pack.get_our().hp <= 0:
                self.force_end = True

        self.force_end = False
        self.register(new_buff(self, Trigger.TAKEN_DAMAGE).name("死亡检测").handler(check_death)
                      .priority(BuffPriority.ABS_FINAL))

    # def add_hook(self, checker, ):
    #     self.hooks.append(checker)



import math
import random

from .tools.stream import Stream
from .logger import Logger
from .msgManager import MsgManager
from .msgPack import MsgPack
from .tools.tools import get_num
from .enums import Trigger, DamageType, BuffTag, BuffPriority
from .buff import new_buff
import uuid

from .gameBoard import GameBoard


class Pokemon:
    def __init__(self, msg_manager: MsgManager, logger: Logger, game_board: GameBoard):
        self.gameBoard: GameBoard = game_board
        self.logger: Logger = logger
        self.msg_manager: MsgManager = msg_manager
        self.name = "÷生"
        self.ATK = 20
        self.DEF = 10
        self.hp = 100  # 血量（随着战斗变化）
        self.MAX_HP = 100  # 血量上限
        self.SPD = 10  # 速度
        self.id = uuid.uuid1()
        self.lv = 1
        # bashlv: 基础等级
        # 指明上面的数据是什么等级的
        # 因为每升1级，三维都会+10%，lv是5 baselv是1 实际战斗攻击力就是20*1.1^(lv-baselv)
        # baselv可以方便填表，不一定要填1级的数据
        # 基础等级是隐藏的，暂时没有修改隐藏等级的词条
        self.baselv = 1
        # fakebaselv: 虽然baselv是5但是成长类还是会从fakebaselv开始计算，方便填表用
        # baselv:影响基础成长
        # fakebaselv:影响成长技能
        self.fakebaselv = 1
        # 这两个外部不需要赋值
        self.CRIT = 0  # 暴击率
        self.CSD = 2  # 暴击伤害

        self._max_hp = None  # 最大血量的缓存，仅仅用于显示血量时不要计算太多次

        self.skillGroup = []
        self.tag: dict[str, int] = {}  # 存放战斗时的一些buff
        self.party: str = ""  # 阵营

    def __str__(self) -> str:
        return f"【{self.name} 技能列表:{self.skillGroup}】"

    def __eq__(self, other):
        if self is None or other is None:  # 无源伤害
            return False
        return self.id == other.id

    def read_pokemon_data(self):
        return

    # 角色tag操作
    def add_tag(self, tag: str, num=1):
        self.tag[tag] = num

    def remove_tag(self, tag: str):
        if tag in self.tag:
            del self.tag[tag]

    def check_tag(self, tag: str):
        return tag in self.tag

    def get_tag(self, tag: str):
        return self.tag[tag]

    def change_tag(self, tag: str, num):
        self.tag[tag] = num

    @staticmethod
    def _attack(self, pack: MsgPack):
        our = pack.get_our()
        enemy = pack.get_enemy()
        # 基本伤害公式
        if pack.is_perfw():  # 结算穿透伤害
            tmp = our.get_atk()
        else:
            tmp = our.get_atk() - enemy.get_def()
        if tmp < 0:
            tmp = 0
        percent = pack.get_percent()
        dmg = tmp * percent // 100

        # 发送即将造成伤害的包，以此计算属性增减伤等属性
        pack2 = MsgPack.damage_pack(our, pack.get_enemy(), dmg, DamageType.NORMAL)
        self.msg_manager.send_msg(pack2)
        dmg = pack2.get_damage()
        # 暴击结算（暂且放在外面,后面复杂了可能移动到buff系统里面）
        crit = False
        if random.random() < self.get_crit() / 100:
            dmg = round(dmg * self.get_csd())
            crit = True
        enemy.hp = round(enemy.hp - dmg)
        element = "毒" if pack2.check_damage_type(DamageType.POISON) else ""
        if crit:
            msg = f"暴击💥！{our.name}对{enemy.name}造成{dmg}点{element}伤害{enemy.get_life_str()}"
        else:
            msg = f"{our.name}对{enemy.name}造成{dmg}点{element}伤害{enemy.get_life_str()}"
        self.logger.log(msg)
        # 被打了，再发一个包
        pack3 = MsgPack.taken_damage_pack(pack.get_enemy(), our, pack2.get_damage(), Trigger.ATTACK, DamageType.NORMAL)
        self.msg_manager.send_msg(pack3)

        if enemy.hp <= 0:
            return True
        return False

    @staticmethod  # 反击逻辑结算
    def counter(self, pack: MsgPack):
        our = pack.get_our()
        enemy = pack.get_enemy()
        # 基本伤害公式
        tmp = max(0, our.get_atk())
        percent = pack.get_percent()
        dmg = tmp * percent // 100

        # 发送即将造成伤害的包，以此计算属性增减伤等属性
        pack2 = MsgPack.damage_pack(our, pack.get_enemy(), dmg, DamageType.REAL)
        self.msg_manager.send_msg(pack2)
        dmg = pack2.get_damage()

        enemy.hp = round(enemy.hp - dmg)
        element = "毒" if pack2.check_damage_type(DamageType.POISON) else ""
        msg = f"{our.name}对{enemy.name}造成{dmg}点{element}伤害{enemy.get_life_str()}"
        self.logger.log(msg)
        # 被打了，再发一个包
        pack3 = MsgPack.taken_damage_pack(pack.get_enemy(), our, pack2.get_damage(), Trigger.COUNTER, DamageType.REAL)
        self.msg_manager.send_msg(pack3)

        if enemy.hp <= 0:
            return True
        return False

    def init(self):
        # 为每个角色注册普攻技能 普攻每个人都有
        def attack_handle(pack: MsgPack):
            Pokemon._attack(self, pack)
            return

        self.msg_manager.register(
            new_buff(self, Trigger.ATTACK)
            .name(f"{self.name}的【普攻】")
            .checker(is_self())
            .handler(attack_handle))

        # 等级对数值的影响（每升1级 1.1倍）
        self.change_atk("等级", multi_eval(lambda: math.pow(1.1, self.get_lv() - self.baselv)))
        self.change_def("等级", multi_eval(lambda: math.pow(1.1, self.get_lv() - self.baselv)))
        self.change_hp("等级", multi_eval(lambda: math.pow(1.1, self.get_lv() - self.baselv)))

        # 为每个角色注册自己的独有技能
        Stream(self.skillGroup).for_each(lambda skill: self.init_skill(skill))
        self.hp = self.get_max_hp()  # 初始化生命值为最大生命值

    def attack(self, enemy: "Pokemon"):
        self.logger.log(f"{self.name}的攻击")
        self.msg_manager.send_msg(MsgPack.atk_pack().our(self).enemy(enemy))

    def get_atk(self):
        pack = MsgPack.get_atk_pack().atk(self.ATK).our(self)
        self.msg_manager.send_msg(pack)
        return round(pack.get_atk())

    def get_def(self):
        pack = MsgPack.get_def_pack().defe(self.DEF).our(self)
        self.msg_manager.send_msg(pack)
        return round(pack.get_def())

    def get_lv(self):
        pack = MsgPack.get_lv_pack().lv(self.lv).our(self)
        self.msg_manager.send_msg(pack)
        return int(pack.get_lv())

    def get_max_hp(self):
        # if self._max_hp is None:
        pack = MsgPack.get_max_hp_pack().hp(self.MAX_HP).our(self)
        self.msg_manager.send_msg(pack)
        self._max_hp = round(pack.get_max_hp())
        if self.hp > self._max_hp:
            self.hp = self._max_hp
        return self._max_hp

    def get_spd(self):
        pack = MsgPack.get_spd_pack().spd(self.SPD).our(self)
        self.msg_manager.send_msg(pack)
        return round(pack.get_spd())

    def get_life_inc_spd(self):
        pack = MsgPack.get_life_inc_spd_pack().life_inc_spd(1).our(self)
        self.msg_manager.send_msg(pack)
        return pack.get_life_inc_spd()

    # 获取当前暴击率
    def get_crit(self):
        pack = MsgPack.get_crit_pack().crit(self.CRIT).our(self)
        self.msg_manager.send_msg(pack)
        return pack.get_crit()

    def get_csd(self):
        pack = MsgPack.get_csd_pack().csd(self.CSD).our(self)
        self.msg_manager.send_msg(pack)
        return pack.get_csd()

    def get_life_str(self):
        tmp = '▁▂▃▄▅▆▇█💀'
        hp = self.hp
        max_hp = self._max_hp
        percent = tmp[max(-1, hp * 7 // max_hp)]
        party = '👿' if self.party != "our" else '🐔'
        return f"\n({party}{percent}{hp}/{max_hp})"

    def init_skill(self, skill: str):
        num = get_num(skill)

        if skill.startswith("利爪"):
            self.logger.log(f"{self.name}的【利爪】发动了！攻击力增加了{num}%")
            self.change_atk(skill, add_percent(num))
            return

        if skill.startswith("尖角"):
            self.logger.log(f"{self.name}的【尖角】发动了！攻击力增加了{num}点")
            self.change_atk(skill, add_num(num))
            return

        if skill.startswith("鳞片"):
            self.logger.log(f"{self.name}的【鳞片】发动了！防御力增加了{num}点")
            self.change_def(skill, add_num(num))
            return

        if skill.startswith("铠甲"):
            self.logger.log(f"{self.name}的【铠甲】发动了！防御力增加了{num}%")
            self.change_def(skill, add_percent(num))
            return

        if skill.startswith("铁壁"):
            self.logger.log(f"{self.name}的【铁壁】发动了！攻击力改为当前防御力的数值")

            def change_atk(pack):
                return self.get_def()

            self.change_atk(skill, change_atk)
            return

        if skill.startswith("机敏"):
            self.logger.log(f"{self.name}的【机敏】发动了！速度增加了{num}点")
            self.change_spd(skill, add_num(num))
            return

        if skill.startswith("舍身"):
            self.logger.log(f"{self.name}的【舍身】发动了！攻击+{num}点，防御-10")
            self.change_atk(skill, add_num(num))
            self.change_def(skill, add_num(-10))
            return

        if skill.startswith("健壮"):
            self.logger.log(f"{self.name}的【健壮】发动了！生命增加了{num}点")
            self.change_hp(skill, add_num(num))
            return

        if skill.startswith("结实"):
            self.logger.log(f"{self.name}的【结实】发动了！生命增加了{num}%")
            self.change_hp(skill, add_percent(num))
            return

        if skill.startswith("毒液"):  # 攻击无法直接造成伤害，改为造成x%的真实伤害，持续2回合
            self.logger.log(f"{self.name}的【毒液】发动了！攻击方式变成{num}%持续2回合的真实伤害")

            def disable_normal_atk(pack: MsgPack):
                p: MsgPack = pack.get_pack()
                return pack.get_owner() == p.get_owner() and p.check_trigger(Trigger.ATTACK) and p.check_buff_name(
                    "【普攻】")

            def poison(pack: MsgPack):
                pack.not_allow()
                our: "Pokemon" = pack.get_pack().get_our()
                enemy: "Pokemon" = pack.get_pack().get_enemy()
                damage = our.get_atk() * num // 100  # 注意中毒计算的atk以施加毒的回合为准
                self.logger.log(f"{our.name}的攻击！{enemy.name}中毒了！受到每回合{damage}点的伤害（持续2回合）")

                def _(pack: MsgPack):
                    # 毒 真无源伤害
                    pack2 = MsgPack.damage_pack(None, pack.get_owner(), damage, DamageType.POISON)
                    self.msg_manager.send_msg(pack2)
                    enemy.hp = round(enemy.hp - pack2.get_damage())
                    self.logger.log(f"{enemy.name}中毒了，流失了{pack2.get_damage()}点血量{pack.get_owner().get_life_str()}")
                    # 受到伤害后发包，死亡结算也在这里进行，所以就算是无源伤害（被毒死）也得发
                    pack3 = MsgPack.taken_damage_pack(pack.get_owner(), None, pack2.get_damage(), Trigger.TURN_END,
                                                      DamageType.POISON)
                    self.msg_manager.send_msg(pack3)

                # 毒buff会直接挂载在敌人身上（无源伤害），taken_damage_pack包的enemy参数会被设为空
                # 哪怕我方有无法造成伤害的debuff，毒也能正常工作
                self.msg_manager.register(new_buff(self, Trigger.TURN_END).owner(enemy).name("【中毒】").
                                          tag(BuffTag.POISON_DEBUFF).handler(_).time(2))

            self.msg_manager.register(
                new_buff(self, Trigger.ACTIVE).name(skill).checker(disable_normal_atk).handler(
                    poison))
            return

        if skill.startswith("连击"):
            self.logger.log(f"{self.name}的【连击】发动了！攻击力变为{num}%，但是每回合攻击2次！")
            # 攻击力大幅降低
            self.change_atk(skill, multi(num / 100))

            # 每回合攻击2次
            def attack_handle(pack: MsgPack):
                self.logger.log(f"{pack.get_owner().name}追加连击")
                Pokemon._attack(self, pack)

            self.msg_manager.register(
                new_buff(self, Trigger.ATTACK).name("【普攻】").checker(is_self()).handler(attack_handle))
            return

        if skill.startswith("反击"):
            self.logger.log(f"{self.name}的【反击】发动了！受到攻击造成的伤害时会以真实伤害反击，反击伤害为正常攻击的{num}%")

            def _handle(pack: MsgPack):
                pack.perfw().percent(num)
                self.logger.log(f"{pack.get_owner().name}的反击！")
                Pokemon.counter(self, pack)

            self.msg_manager.register(
                new_buff(self, Trigger.TAKEN_DAMAGE).name("【反击】").checker(is_self())
                .checker(lambda pack: pack.check_damage_taken_trigger(Trigger.ATTACK)).handler(_handle))
            return

        if skill.startswith("不屈"):
            self.logger.log(f"{self.name}的【不屈】发动了！最大生命值增加{num}%")
            self.msg_manager.register(
                new_buff(self, Trigger.GET_HP).name(skill).checker(is_self()).handler(
                    lambda pack: pack.change_max_hp(add_percent(num))))
            return

        if skill.startswith("野性"):
            self.logger.log(f"{self.name}的【野性】发动了！最终伤害增加+{num}%")
            self.change_damage("野性", add_percent(num))
            return
        if skill.startswith("尖牙"):
            self.logger.log(f"{self.name}的【尖牙】发动了！暴击率+{num}%")
            self.change_crit("尖牙", add_num(num))
            return
        if skill.startswith("坚韧"):
            self.logger.log(f"{self.name}的【坚韧】发动了！最大防御力增加{num}%")
            self.msg_manager.register(
                new_buff(self, Trigger.GET_DEF).name(skill).checker(is_self()).handler(
                    lambda pack: pack.change_def(add_percent(num))))
            return

        if skill.startswith("愈合"):
            self.logger.log(f"{self.name}发动【愈合】!回合末回复{num}%,未受击中伤再多回复一半")

            TAG = "受到了伤害"

            def gain_hp(pack):
                max_hp = self.get_max_hp()
                hp_gain = (1 if self.check_tag(TAG) else 1.5) * max_hp * num / 100
                hp_gained = round(min(hp_gain * self.get_life_inc_spd(), max_hp - self.hp))
                self.hp = round(hp_gained + self.hp)
                self.logger.log(f"{pack.get_owner().name}的【愈合】发动了！{('' if self.check_tag(TAG) else '充分')}"
                                f"回复了{hp_gained}hp{pack.get_owner().get_life_str()}")

            self.msg_manager.register(
                new_buff(self, Trigger.TURN_END).name(skill).handler(gain_hp))
            self.msg_manager.register(
                new_buff(self, Trigger.TAKEN_DAMAGE).name("愈合检测受伤").checker(is_self()).
                checker(lambda pack: pack.get_damage() > 0).checker(
                    lambda pack: pack.check_damage_taken_trigger(Trigger.ATTACK))
                .handler(
                    lambda pack: self.add_tag(TAG)
                )
            )
            self.msg_manager.register(new_buff(self, Trigger.TURN_END).name("愈合重置检测").handler(
                lambda pack: self.remove_tag(TAG)
            ))
            return

        if skill.startswith("长生"):
            self.logger.log(f"{self.name}的【长生】发动了！生命回复速+{num}%")
            self.msg_manager.register(
                new_buff(self, Trigger.GET_LIFE_INC_SPD).name(skill).checker(is_self()).handler(
                    lambda pack: pack.change_life_inc_spd(multi(2))))
            return

        if skill.startswith("攻击成长"):
            self.logger.log(f"{self.name}的【攻击成长】使其攻击增加了{self.get_lv() - self.fakebaselv}*{num}点")
            self.change_atk(skill, add_num_eval(lambda: num * (self.get_lv() - self.fakebaselv)))
            return

        if skill.startswith("防御成长"):
            self.logger.log(f"{self.name}的【防御成长】使其防御增加了{self.get_lv() - self.fakebaselv}*{num}点")
            self.change_def(skill,
                            add_num_eval(lambda: num * (self.get_lv() - self.fakebaselv)))
            return

        if skill.startswith("生命成长"):
            self.logger.log(f"{self.name}的【生命成长】使其生命增加了{self.get_lv() - self.fakebaselv}*{num}点")
            self.change_hp(skill, add_num_eval(lambda: num * (self.get_lv() - self.fakebaselv)))
            return

        if skill.startswith("地区霸主"):
            self.logger.log(f"看起来这片区域的主人出现了...")
            self.change_lv(skill, add_num(1))
            # 技能槽位的增加目前仅仅存在于设定之中，怪物的技能数量其实是填表决定的，并非靠这个技能获得
            return

        if skill.startswith("天才"):
            self.logger.log(f"{self.name}的天才使其等级上升了{num}")
            self.change_lv(skill, add_num(num))
            # self.fakebaselv -= 1
            return

        if skill.startswith("剧毒之体"):
            self.logger.log(f"{self.name}发动【剧毒之体】，以不能造成毒以外伤害的代价换来敌人受到毒伤+20%，回合末立刻解毒")
            # 造成的毒伤害加倍 非毒伤害归零
            self.msg_manager.register(
                new_buff(self, Trigger.DEAL_DAMAGE).name(skill).checker(is_enemy_taken(self)).priority(
                    BuffPriority.CHANGE_DAMAGE_LAST).handler(
                    lambda pack: pack.damage(
                        pack.get_damage() * 1.2 if pack.check_damage_type(DamageType.POISON) else 0)))

            # 回合结束时，强制解除所有中毒debuff

            def del_poison_buff(pack: MsgPack):
                self.msg_manager.get_buff_stream() \
                    .filter(lambda b: b.check_tag(BuffTag.POISON_DEBUFF)) \
                    .filter(lambda b: b.check_owner(self)) \
                    .for_each(lambda b: (b.mark_as_delete(), self.logger.log(f"{self.name}解除了自己中的毒")))

            self.msg_manager.register(
                new_buff(self, Trigger.TURN_END).name("剧毒之体解毒").handler(lambda pack: del_poison_buff(pack)).priority(
                    BuffPriority.TURN_END_LAST))

            return

        if skill.startswith("猛毒"):
            self.logger.log(f"{self.name}的猛毒发动了！敌人受到毒伤害+{num}%")
            self.msg_manager.register(
                new_buff(self, Trigger.DEAL_DAMAGE).name(skill).checker(is_enemy_taken(self))
                .checker(lambda pack: pack.check_damage_type(DamageType.POISON))
                .handler(lambda pack: pack.damage(pack.get_damage() * (100 + num) / 100)))
            return

        if skill.startswith("毒刃"):
            self.logger.log(f"{self.name}的毒刃发动了！攻击伤害增加{num}%然后变为毒属性")
            self.change_damage("毒刃", add_percent(num))
            self.msg_manager.register(
                new_buff(self, Trigger.DEAL_DAMAGE).name(skill).checker(is_self())
                .handler(lambda pack: pack.damage_type(DamageType.POISON)))
            return

        if skill.startswith("暴怒"):
            self.logger.log(f"{self.name}的暴怒发动了！受伤越重攻击越高，最高增加{num}！")

            def change_atk_by_anger(pack: MsgPack):
                our: Pokemon = pack.get_our()
                max_hp = our.get_max_hp()
                hp = pack.get_our().hp
                percent = (1 - hp / max_hp) * num
                pack.change_atk(add_num(percent))

            self.msg_manager.register(
                new_buff(self, Trigger.GET_ATK).name(skill).checker(is_self()).handler(
                    lambda pack: change_atk_by_anger(pack)))
            return

        if skill.startswith("惜别"):
            TAG = "惜别已使用"
            self.logger.log(f"{self.name}的惜别发动了，可以1血复活一次")

            def tear(pack: MsgPack):
                our: Pokemon = pack.get_our()
                if not our.check_tag(TAG) and our.hp <= 0:
                    our.hp = 1
                    our.add_tag(TAG)
                    self.logger.log(f"濒死之际，一股意志支撑{our.name}又活了过来{our.get_life_str()}")
                    self.msg_manager.get_buff_stream() \
                        .filter(lambda b: b.check_tag(BuffTag.POISON_DEBUFF)) \
                        .filter(lambda b: b.check_owner(self)) \
                        .for_each(
                        lambda b: (b.mark_as_delete().mark_as_disable(), self.logger.log(f"{self.name}解除了自己中的毒")))

            self.msg_manager.register(
                new_buff(self, Trigger.TAKEN_DAMAGE).name("惜别").checker(is_self()).handler(
                    lambda pack: tear(pack)
                )
            )

            return

        if skill.startswith("抗毒"):
            self.logger.log(f"{self.name}对毒有抵抗力，受到毒伤害降低{num}%")
            self.msg_manager.register(
                new_buff(self, Trigger.DEAL_DAMAGE).name("抗毒").checker(is_enemy())
                .priority(BuffPriority.CHANGE_DAMAGE_LAST)
                .checker(lambda pack: pack.check_damage_type(DamageType.POISON))
                .handler(lambda pack: pack.damage(pack.get_damage() * (100 - num) / 100)))
            return

        if skill.startswith("吸收"):
            # todo
            self.logger.log(f"{self.name}发动【吸收】,+{num}攻击+{num}%攻击吸血")
            # 攻击增加
            self.change_atk(skill, add_num(num))

            def life_drain(pack: MsgPack):
                if pack.check_damage_taken_trigger(Trigger.ATTACK):
                    drain = round(pack.get_damage() * num * self.get_life_inc_spd() / 100)
                    self.hp = min(self.get_max_hp(), round(self.hp + drain))
                    self.logger.log(f"{self.name}吸取了{drain}点血量{self.get_life_str()}")

            # 吸血
            self.msg_manager.register(new_buff(self, Trigger.TAKEN_DAMAGE).checker(is_enemy()).handler(life_drain))
            return

        # 这三个强关联技能违反隔离原则一起写了，也可以分开写
        if skill.startswith("游泳"):
            self.logger.log(f"{self.name}的【游泳】发动了，能憋气{num}回合")
            if self.check_tag("溺水"):
                self.change_tag("溺水", self.get_tag("溺水") - num)
            else:
                self.add_tag("溺水", -num)
            return

        if skill.startswith("溺水"):
            self.logger.log(f"{self.name}来到了水中...")

            def handle_swim(pack):
                turn = self.get_tag("溺水")
                if turn < 0:
                    self.logger.log(f"{self.name}的氧气还能坚持{-turn}回合")
                else:
                    self.logger.log(f"{self.name}喘不上气了！溺水:{turn + 1}☠")
                self.change_tag("溺水", turn + 1)

            def handle_swim2(pack):
                turn = self.get_tag("溺水")
                if turn > 0:
                    # 发送即将造成伤害的包，以此计算属性增减伤等属性
                    pack2 = MsgPack.damage_pack(None, self, self.get_max_hp() * 0.03 * turn, DamageType.REAL)
                    self.msg_manager.send_msg(pack2)
                    self.hp = round(self.hp - pack2.get_damage())
                    # 受到伤害后发包，死亡结算也在这里进行，所以就算是无源伤害（被毒死）也得发
                    pack3 = MsgPack.taken_damage_pack(self, None, pack2.get_damage(), Trigger.TURN_START,
                                                      DamageType.REAL)
                    self.msg_manager.send_msg(pack3)
                    self.logger.log(f"{self.name}在水中挣扎！流失了{pack2.get_damage()}点血量{pack.get_owner().get_life_str()}")

            if not self.check_tag("溺水"):
                self.add_tag("溺水", 0)
            self.msg_manager.register(
                new_buff(self, Trigger.TURN_START).name("【溺水】")
                .priority(BuffPriority.ENVIRONMENT)
                .handler(handle_swim))
            self.change_damage("【溺水】减攻", add_percent_eval(lambda: round(max(self.get_tag("溺水"), 0) * -5, 2)))
            self.msg_manager.register(
                new_buff(self, Trigger.TURN_START).name("【溺水】体力流失")
                .handler(handle_swim2))
            return

        if skill.startswith("水栖"):
            self.logger.log(f"{self.name}是水中的生物")

            # 免疫溺水
            def get_status():
                if self.check_tag("溺水") and self.get_tag("溺水") >= 0:
                    return False  # 处于溺水中
                return True

            def disable_water(pack: MsgPack):
                p: MsgPack = pack.get_pack()
                return pack.get_owner() == p.get_owner() and p.check_buff_name("【溺水】")

            self.change_damage("不溺水减伤害", add_percent_eval(lambda: -50 if get_status() else 0))
            self.change_damage("溺水速度翻倍", multi_eval(lambda: 2 if get_status() else 1))
            self.msg_manager.register(
                new_buff(self, Trigger.ACTIVE).name(skill).checker(disable_water).handler(
                    lambda pack: pack.not_allow()))
            self.change_damage("免疫溺水伤害", multi_eval(lambda: 2 if get_status() else 1))

            # todo  场地额外副作用
            if "溺水" not in self.skillGroup:
                self.init_skill("溺水")
            if "溺水" not in self.gameBoard.enemy.skillGroup:
                self.gameBoard.our.init_skill("溺水")

            return

        if skill.startswith("灵巧"):
            self.logger.log(f"{self.name}的【灵巧】发动了，获得{num}%的闪避")

            def avoid_atk(pack: MsgPack):
                if pack.get_pack().get_enemy() == self and random.random() <= num / 100:
                    pack.not_allow()
                    self.logger.log(f"{self.name}灵巧的躲过了攻击")
                    return True
                return False

            self.msg_manager.register(
                new_buff(self, Trigger.ACTIVE).name(skill).checker(
                    lambda pack: pack.get_pack().check_trigger(Trigger.ATTACK))
                .checker(avoid_atk).handler(lambda pack: None))
            return

        if skill.startswith("粘液"):
            if self.check_tag("蛊神"):
                num = num * 2
            self.logger.log(f"{self.name}用滑溜溜的【粘液】降低了敌人{num}%攻击力")
            self.msg_manager.register(
                new_buff(self, Trigger.GET_ATK).name(skill).checker(is_enemy()).handler(
                    lambda pack: pack.change_atk(add_percent(-num))).priority(BuffPriority.CHANGE_ATK_LAST))
            return

        if skill.startswith("咒眼"):
            if self.check_tag("蛊神"):
                num = num * 2
            self.logger.log(f"{self.name}用【咒眼】凝视着敌人，降低敌人{num}基础攻击力")
            self.msg_manager.register(
                new_buff(self, Trigger.GET_ATK).name(skill).checker(is_enemy()).handler(
                    lambda pack: pack.change_atk(add_num(-num))).priority(BuffPriority.CHANGE_ATK_FIRST))
            return
        if skill.startswith("铁壁"):
            self.logger.log(f"{self.name}的铁壁发动了！攻击力变成与防御力相同")

            def replace_atk_to_def(atk):
                return self.get_def()

            self.msg_manager.register(
                new_buff(self, Trigger.GET_ATK).name(skill).checker(is_self()).handler(
                    lambda pack: pack.change_atk(replace_atk_to_def)))
            return
        if skill.startswith("蛊神"):
            self.logger.log(f"{self.name}的【蛊神】发动了！所有诅咒效果变成2倍")
            # 懒得用触发器写了，直接用tag
            self.add_tag("蛊神")
            return
        if skill.startswith("威压"):
            num = 1
            if self.check_tag("蛊神"):
                num = num * 2
            self.logger.log(f"{self.name}的【威压】让敌人下降了{num}级")
            self.msg_manager.register(
                new_buff(self, Trigger.GET_LV).name(skill).checker(is_enemy()).handler(
                    lambda pack: pack.change_lv(add_num(-num))))
            return

        if skill.startswith("伤噬"):
            if self.check_tag("蛊神"):
                num = num * 2
            self.logger.log(f"{self.name}的【伤噬】让敌人受伤+{num}%")

            self.msg_manager.register(
                new_buff(self, Trigger.DEAL_DAMAGE).name(skill).checker(is_enemy_taken(self))
                .handler(lambda pack: pack.damage(pack.get_damage() * (100 + num) / 100)))
            return

        raise Exception(f"不认识的技能：{skill}")

    def change_atk(self: "Pokemon", skill, func):
        self.msg_manager.register(
            new_buff(self, Trigger.GET_ATK).name(skill).checker(is_self()).handler(
                lambda pack: pack.change_atk(func)))

    def change_def(self: "Pokemon", skill, func):
        self.msg_manager.register(
            new_buff(self, Trigger.GET_DEF).name(skill).checker(is_self()).handler(
                lambda pack: pack.change_def(func)))

    def change_hp(self: "Pokemon", skill, func):
        self.msg_manager.register(
            new_buff(self, Trigger.GET_HP).name(skill).checker(is_self()).handler(
                lambda pack: pack.change_max_hp(func)))

    def change_lv(self: "Pokemon", skill, func):
        self.msg_manager.register(
            new_buff(self, Trigger.GET_LV).name(skill).checker(is_self()).handler(
                lambda pack: pack.change_lv(func)))

    def change_damage(self, skill, func):
        self.msg_manager.register(
            new_buff(self, Trigger.DEAL_DAMAGE).name(skill).checker(is_self())
            .handler(lambda pack: pack.change_damage(func)))

    def change_spd(self, skill, func):
        self.msg_manager.register(
            new_buff(self, Trigger.GET_SPD).name(skill).checker(is_self())
            .handler(lambda pack: pack.change_spd(func)))

    def change_crit(self, skill, func):
        self.msg_manager.register(
            new_buff(self, Trigger.GET_CRIT).name(skill).checker(is_self())
            .handler(lambda pack: pack.change_crit(func)))

    def change_csd(self, skill, func):
        self.msg_manager.register(
            new_buff(self, Trigger.GET_CSD).name(skill).checker(is_self())
            .handler(lambda pack: pack.change_crit(func)))


#
# def check_id(name):
#     def _(pack: MsgPack):
#         return pack.check_name(name)
#
#     return _

def is_self():
    def _(pack: MsgPack):
        return pack.is_our_owner()

    return _


def is_enemy():
    def _(pack: MsgPack):
        return not pack.is_our_owner()

    return _


def is_enemy_taken(self):
    def _(pack: MsgPack):
        return pack.get_enemy() != self

    return _


# 旧写法已废弃
# def is_enemy2(): # 这个方法只能用于攻击 获取自身数据（如攻击力时）没有敌人对象用不了
#     def _(pack: MsgPack):
#         return pack.is_enemy_owner()
#
#     return _

def is_pokemon(p: Pokemon):
    def _(pack: MsgPack):
        return pack.get_our() == p

    return _


def add_num(num):
    def _(origin):
        return origin + num

    return _


def add_num_eval(func):
    def _(origin):
        return origin + func()

    return _


def multi(num):
    def _(origin):
        return origin * num

    return _


def multi_eval(func):
    def _(origin):
        return origin * func()

    return _


def add_percent(num):
    def _(origin):
        return origin * (1 + num / 100)

    return _


def add_percent_eval(func):
    def _(origin):
        # print(func())
        return origin * (1 + func() / 100)

    return _



import io


class Logger:
    _log = io.StringIO()
    _debug = False

    def log(self, string):
        if self._debug:
            print(string)
        else:
            self._log.write(string)
            self._log.write('\n')

    def print_log(self):
        print(self._log.getvalue())

    def clean(self):
        self._log = io.StringIO()
        pass

    def get_log(self):
        # print(self._log)
        return self._log.getvalue()

    def debug_log(self, string):
        print(string)



import time

from nonebot import logger
from nonebot.adapters.onebot.v11 import GroupMessageEvent

from ..psql_db import db
from ..service.adv_handler import get_user_status
from ..model.player_model import PlayerDB
from ..utils import get_uid


async def register_new_player(event: GroupMessageEvent, name: str, head_choose: str) -> str:
    player = await PlayerDB.get_player_by_name(name)
    if player:
        return "这个名字已经有人用过啦！"

    player_count = await PlayerDB.register(get_uid(event), event.group_id, name, head_choose)

    if player_count:
        return f"{name}加入了只因进化录的世界！当前赛季为S1起源赛季，你是第{player_count}只因哦\n" \
               f"输入【领取任务】指令开始你的新手引导吧！"
    else:
        return "未知错误..."


async def get_player(event) -> PlayerDB:
    player = await PlayerDB.get_player_by_uid(get_uid(event))
    return player


async def get_user_status_str(event) -> str:
    player, status, pos_list = await get_user_status(event)
    return f"{player.name} lv{player.system_lv}{status}\n" \
           f"所有掌握的技能：\n" \
           f"{player.show_skill()}\n" \
           f"背包：\n" \
           f"{player.show_bag()}"


async def handle_set_stay(event, lv: int) -> str:
    player = await PlayerDB.get_player_by_uid(get_uid(event))
    if lv <= 0:
        if "stay" in player.buff:
            del player.buff["stay"]
            await player.update(buff=player.buff).apply()
        return "已经解除驻留功能"

    player.buff["stay"] = lv + 1
    s = f"驻留功能已开启，角色遇到lv.{lv}的怪物后不再前往下一区域"
    await player.update(buff=player.buff).apply()
    return s


async def handle_show_bag(event) -> str:
    player = await PlayerDB.get_player_by_uid(get_uid(event))
    return player.show_bag()


async def reward_all() -> str:
    # todo 性能优化+ 限制只给对应赛季发放
    players: list[PlayerDB] = await PlayerDB.get_all()
    try:
        async with db.transaction():
            for player in players:
                player.add_item("小沙漏", 1)
                await player.update(bag=player.bag, collection=player.collection).apply()
    except Exception as e:
        logger.info(f"全服发放奖励出错 {type(e)}: {e}")
        return "全服发放奖励出错"
    return "全服发放奖励 小沙漏*1 成功"


async def query_times(event) -> int:
    player = await PlayerDB.get_player_by_uid(get_uid(event))
    return player.times


async def change_name(event, name) -> str:
    if len(name) > 10:
        return "名字不能超过10个字符"
    if len(name) == 0:
        return "名字太短了！"
    player = await PlayerDB.get_player_by_uid(get_uid(event))
    await player.update(name=name).apply()
    return f"{name},名称修改成功！"


async def auto_sign(player: PlayerDB) -> str:
    # 自动签到逻辑
    now_day = time.strftime('%Y-%m-%d')
    if player.buff.get("sign_day", "") != now_day:
        player.add_item("小沙漏", 1)
        player.buff["sign_day"] = now_day
        await player.update(bag=player.bag, collection=player.collection, buff=player.buff).apply()
        return "今日签到已完成，获得小沙漏一个喵"
    return ""



from ..stream import Stream
from ..utils import add_item, add_items
from nonebot import logger
from ..psql_db import db

from datetime import datetime
from nonebot import logger



class PlayerDB(db.Model):
    __tablename__ = "ikun_player"

    id: int = db.Column(db.Integer(), primary_key=True)
    # 角色ID
    uid: str = db.Column(db.String(), nullable=False)
    # 角色名字
    name: str = db.Column(db.String(), nullable=False)
    # 攻击力
    atk: str = db.Column(db.Integer(), nullable=False)
    # 防御力
    defe: str = db.Column(db.Integer(), nullable=False)
    # 生命值
    hp: str = db.Column(db.Integer(), nullable=False)
    # 速度
    spd: str = db.Column(db.Integer(), nullable=False)
    # 头像
    head: str = db.Column(db.String(), nullable=False)
    # 探险等级（每去过一个地方，等级就会+1）
    adv_lv: int = db.Column(db.Integer(), nullable=False)
    # 系统等级（影响可同时装备的技能数量）
    system_lv: int = db.Column(db.Integer(), default=1)
    # 角色等级
    lv: int = db.Column(db.Integer(), nullable=False)
    # 已装备技能
    equip_skill: list[str] = db.Column(db.JSON(), nullable=False, default=[])
    # 技能
    skill = db.Column(db.JSON(), nullable=False, default={})
    # 体质
    sp_skill = db.Column(db.JSON(), nullable=False, default={})
    # 背包
    bag: dict[str, int] = db.Column(db.JSON(), nullable=False, default={'IKUN图册': 1})
    # buff
    buff = db.Column(db.JSON(), nullable=False, default={})
    # 箱子
    box = db.Column(db.JSON(), nullable=False, default={})
    # 击杀怪物的次数（起名是历史问题）
    monster_killed = db.Column(db.JSON(), nullable=False, default={})
    # 遇见怪物的次数
    monster_meet = db.Column(db.JSON(), nullable=False, default={})
    # 游历地点收集
    arrived = db.Column(db.JSON(), nullable=False, default={})
    # 角色标签
    tag = db.Column(db.JSON(), nullable=False, default={})
    # 角色称号
    honor = db.Column(db.JSON(), nullable=False, default={})
    # 出生时间
    create_time = db.Column(db.DateTime(), default=datetime.now)
    # 来自群
    group_id = db.Column(db.BigInteger(), default=0)
    # 已经完成的任务
    complete_mission = db.Column(db.JSON(), nullable=False, default={})
    # 游玩了多少个时点
    times = db.Column(db.Integer(), nullable=False, default=0)
    # 图鉴（暂时不在使用）
    collection = db.Column(db.JSON(), nullable=False, default={'IKUN图册': 1})
    # 缺省字段
    reverse1 = db.Column(db.JSON(), nullable=True, default={})
    reverse2 = db.Column(db.JSON(), nullable=True, default={})
    reverse3 = db.Column(db.JSON(), nullable=True, default={})

    @classmethod
    async def get_player_by_name(
            cls,
            name: str,
    ) -> "PlayerDB":
        try:
            async with db.transaction():
                return await cls.query.where(cls.name == name).gino.first()
        except Exception as e:
            logger.info(f"根据角色名字查询数据库出错 {type(e)}: {e}")

    @classmethod
    async def get_player_by_uid(
            cls,
            uid: str,
    ) -> "PlayerDB":
        try:
            async with db.transaction():
                player = await cls.query.where(cls.uid == uid).gino.first()
                if player is None:
                    logger.info(f"未查询到角色 可能是在注册 查询id为{uid}")
                return player
        except Exception as e:
            logger.info(f"根据角色uid查询数据库出错 {type(e)}: {e}")

    @classmethod
    async def register(cls, uid: str, group_id: int, name: str, head_choose: str) -> int:
        try:
            async with db.transaction():
                logger.info(f"{name}加入游戏")
                await cls.create(uid=uid, name=name, head=head_choose, adv_lv=0, system_lv=1, lv=1, group_id=group_id,
                                 atk=20, defe=10, hp=100, spd=0, sp_skill={'吞天者': 1})
                return await db.func.count(cls.id).gino.scalar()  # 虽然不需要，但是应该是具备性能的写法
        except Exception as e:
            logger.info(f"新增角色出错 {type(e)}: {e}")

    @classmethod
    async def set_status(cls, uid: str, status: str) -> None:
        try:
            async with db.transaction():
                s = await cls.get_player_by_uid(uid)
                await s.update(status=status).apply()
        except Exception as e:
            logger.info(f"设置角色状态出错 {type(e)}: {e}")

    @classmethod
    async def get_all(cls) -> list["PlayerDB"]:
        try:
            async with db.transaction():
                return await PlayerDB.query.gino.all()
        except Exception as e:
            logger.info(f"查询所有角色出错 {type(e)}: {e}")

    # 注意 add完仍然需要update哦
    def add_item(self, name, cnt):
        add_item(self.bag, name, cnt)
        add_item(self.collection, name, cnt)

    def add_items(self, items) -> "PlayerDB":
        self.bag = add_items(self.bag, items)
        self.collection = add_items(self.collection, items)
        return self

    def show_bag(self) -> str:
        if len(self.bag) != 0:
            return Stream(self.bag.items()).map(lambda i: f"{i[0]} x{i[1]}\n").reduce(lambda s1, s2: s1 + s2, '')
        return "没有任何物品"

    def show_equip(self) -> str:
        if len(self.equip) != 0:
            return Stream(self.equip.items()).map(lambda i: f"{i[0]}: {i[1]}\n").reduce(lambda s1, s2: s1 + s2, '')
        return "没穿任何装备"

    # 需要在外部检查物品数量是否足够
    def cost_item(self, item_name, num=1):
        if self.bag.get(item_name, 0) >= num:
            self.bag[item_name] -= num
            if self.bag.get(item_name) == 0:
                self.bag.pop(item_name)
            return True
        else:
            logger.warning(f"使用{item_name}时发现物品不足")
            return False

    def cost_items(self, items: dict[str, int]):
        for i in items.items():
            self.cost_item(i[0], i[1])

    def query_item(self, item_name):
        return self.bag.get(item_name, 0)

    def query_arrived(self, name):
        return self.arrived.get(name, 0)

    def query_skill(self, skill_name):
        return self.skill.get(skill_name, 0)

    def query_skill_fuzzy(self, skill_name):
        return Stream(self.skill.keys()).filter(lambda k: skill_name in k).find_first()

    def get_bag(self) -> dict[str, int]:
        return self.bag

    # 装备系统暂时已经删除了
    def get_equip(self, _type) -> str:
        return self.equip.get(_type)

    def wear(self, _type, name):
        self.equip[_type] = name
        self.cost_item(name)

    def unwear(self, _type, name):
        # 这里的添加不会增加图鉴计数，所以不能直接调用playerDB.add_item
        add_item(self.bag, name, 1)
        self.equip.pop(_type)

    def record_monster_killed(self, monster_name):
        add_item(self.monster_killed, monster_name, 1)

    def record_monster_meet(self, monster_name):
        add_item(self.monster_meet, monster_name, 1)
